"""
分类指标分析模块; 用于教学模拟用

给定分类场景和指标分析现实合理性并进行可行解的推断
"""

import typing

import warnings
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
from scipy.optimize import minimize


@dataclass
class ClassificationSpec:
    """分类问题规格"""

    n_samples: int
    n_classes: int
    class_distribution: Optional[List[float]] = None  # 类别分布比例

    # 给定的指标
    accuracy: Optional[float] = None
    precision: Optional[Union[float, List[float]]] = None
    recall: Optional[Union[float, List[float]]] = None
    f1_score: Optional[Union[float, List[float]]] = None
    specificity: Optional[Union[float, List[float]]] = None

    # 随机种子
    random_seed: Optional[int] = None


@dataclass
class ValidationResult:
    """验证结果"""

    is_feasible: bool
    error_message: Optional[str] = None
    inferred_metrics: Optional[Dict[str, float]] = None
    confusion_matrix: Optional[np.ndarray] = None


class ClassificationMetricsValidator:
    """分类指标验证器"""

    def __init__(self, spec: ClassificationSpec):
        self.spec = spec
        self.rng = np.random.RandomState(spec.random_seed)

    def validate_and_infer(self) -> ValidationResult:
        """验证指标可行性并推断缺失指标"""

        # 1. 基础验证
        basic_validation = self._validate_basic_constraints()
        if not basic_validation.is_feasible:
            return basic_validation

        # 2. 设置默认类别分布（避免修改原对象）
        class_dist = self.spec.class_distribution
        if class_dist is None:
            class_dist = [1.0 / self.spec.n_classes] * self.spec.n_classes
        else:
            class_dist = class_dist.copy()  # 创建副本避免修改原对象

        # 3. 归一化类别分布
        total_ratio = sum(class_dist)
        normalized_dist = [r / total_ratio for r in class_dist]

        # 4. 根据类别数选择验证策略
        if self.spec.n_classes == 2:
            return self._validate_binary_classification(normalized_dist)
        else:
            return self._validate_multiclass_classification(normalized_dist)

    def _validate_basic_constraints(self) -> ValidationResult:
        """验证基础约束"""

        if self.spec.n_samples <= 0:
            return ValidationResult(False, '样本数必须为正数')

        if self.spec.n_classes < 2:
            return ValidationResult(False, '类别数必须至少为2')

        # 验证指标范围
        for metric_name, metric_value in [
            ('accuracy', self.spec.accuracy),
            ('precision', self.spec.precision),
            ('recall', self.spec.recall),
            ('f1_score', self.spec.f1_score),
            ('specificity', self.spec.specificity),
        ]:
            if metric_value is not None:
                if isinstance(metric_value, (list, np.ndarray)):
                    if any(v < 0 or v > 1 for v in metric_value):
                        return ValidationResult(False, f'{metric_name}必须在[0,1]范围内')
                else:
                    if metric_value < 0 or metric_value > 1:
                        return ValidationResult(False, f'{metric_name}必须在[0,1]范围内')

        return ValidationResult(True)

    def _validate_binary_classification(
        self, class_distribution: typing.Sequence[float]
    ) -> ValidationResult:
        """验证二分类问题"""

        # 计算类别样本数
        n_pos = round(self.spec.n_samples * class_distribution[1])
        n_neg = self.spec.n_samples - n_pos

        # 检查极端情况
        if n_pos == 0 or n_neg == 0:
            return ValidationResult(
                False, f'类别分布导致某类样本数为0（正类：{n_pos}，负类：{n_neg}）'
            )

        # 收集已知指标
        known_metrics = {}
        for metric_name, metric_value in [
            ('accuracy', self.spec.accuracy),
            ('precision', self.spec.precision),
            ('recall', self.spec.recall),
            ('f1_score', self.spec.f1_score),
            ('specificity', self.spec.specificity),
        ]:
            if metric_value is not None:
                known_metrics[metric_name] = metric_value

        # 验证指标组合的数学可行性
        feasibility_check = self._check_mathematical_feasibility(
            n_pos, n_neg, known_metrics
        )
        if not feasibility_check.is_feasible:
            return feasibility_check

        # 推断混淆矩阵
        try:
            confusion_matrix, inferred_metrics = self._infer_binary_confusion_matrix(
                n_pos, n_neg, known_metrics
            )

            # 验证推断结果的数学一致性
            validation_result = self._validate_inferred_metrics(
                known_metrics, inferred_metrics
            )
            if not validation_result.is_feasible:
                return validation_result

            return ValidationResult(
                True, inferred_metrics=inferred_metrics, confusion_matrix=confusion_matrix
            )

        except Exception as e:
            return ValidationResult(False, f'推断过程出错: {str(e)}')

    def _check_mathematical_feasibility(
        self, n_pos: int, n_neg: int, known_metrics: typing.Mapping[str, float]
    ) -> ValidationResult:
        """检查指标组合的数学可行性"""

        # 如果同时给定了precision, recall, accuracy，需要检查是否兼容
        if all(k in known_metrics for k in ['accuracy', 'precision', 'recall']):
            return self._check_precision_recall_accuracy_compatibility(
                n_pos, n_neg, known_metrics
            )

        # 如果给定了F1和其他指标，检查兼容性
        if 'f1_score' in known_metrics and (
            'precision' in known_metrics or 'recall' in known_metrics
        ):
            return self._check_f1_compatibility(known_metrics)

        return ValidationResult(True)

    def _check_precision_recall_accuracy_compatibility(
        self, n_pos: int, n_neg: int, known_metrics: typing.Mapping[str, float]
    ) -> ValidationResult:
        """检查precision, recall, accuracy三者的兼容性"""

        accuracy = known_metrics['accuracy']
        precision = known_metrics['precision']
        recall = known_metrics['recall']

        # 根据recall计算TP的理论值
        tp_from_recall = recall * n_pos
        fn_from_recall = (1 - recall) * n_pos

        # 根据precision计算FP的理论值
        if precision == 0:
            if tp_from_recall > 0:
                return ValidationResult(False, 'precision=0但recall>0，数学上不可能')
            tp_theoretical = 0
            fp_theoretical = 0
        else:
            tp_theoretical = tp_from_recall
            fp_theoretical = tp_from_recall / precision - tp_from_recall

        # 根据accuracy计算TN的理论值
        tn_from_accuracy = accuracy * (n_pos + n_neg) - tp_theoretical

        # 检查TN + FP = n_neg的约束
        total_neg_predictions = tn_from_accuracy + fp_theoretical

        # 允许一定的数值误差和整数舍入误差
        tolerance = 1.5  # 考虑到整数舍入的误差

        if abs(total_neg_predictions - n_neg) > tolerance:
            return ValidationResult(
                False,
                f'指标组合不可行：根据给定指标推算的负样本数({total_neg_predictions:.2f})与实际负样本数({n_neg})不符，'
                f'差异为{abs(total_neg_predictions - n_neg):.2f}，超过容忍范围{tolerance}',
            )

        # 检查所有值是否为非负
        if any(
            v < -tolerance
            for v in [tp_theoretical, fp_theoretical, tn_from_accuracy, fn_from_recall]
        ):
            return ValidationResult(False, '推导出的混淆矩阵包含负值，指标组合不可行')

        # 检查整数约束的可行性
        # 对于小样本，检查是否能找到合理的整数解
        if n_pos <= 20 or n_neg <= 20:  # 小样本情况
            integer_solution = self._find_integer_solution(
                n_pos, n_neg, accuracy, precision, recall
            )
            if integer_solution is None:
                return ValidationResult(False, '在整数约束下找不到满足所有指标的解')

        return ValidationResult(True)

    def _find_integer_solution(
        self, n_pos: int, n_neg: int, accuracy: float, precision: float, recall: float
    ) -> Optional[Tuple[int, int, int, int]]:
        """尝试找到满足整数约束的解"""

        # 枚举所有可能的TP值
        for tp in range(n_pos + 1):
            fn = n_pos - tp

            # 根据precision计算FP
            if precision == 0:
                if tp > 0:
                    continue
                fp = 0
            else:
                if tp == 0:
                    continue
                fp_exact = tp / precision - tp
                fp = round(fp_exact)
                if fp < 0 or fp > n_neg:
                    continue

                # 检查precision约束
                if abs(tp / (tp + fp) - precision) > 0.1:  # 10%的容忍度
                    continue

            tn = n_neg - fp
            if tn < 0:
                continue

            # 检查recall约束
            if tp + fn > 0:
                calc_recall = tp / (tp + fn)
                if abs(calc_recall - recall) > 0.1:  # 10%的容忍度
                    continue

            # 检查accuracy约束
            calc_accuracy = (tp + tn) / (n_pos + n_neg)
            if abs(calc_accuracy - accuracy) > 0.05:  # 5%的容忍度
                continue

            return (tp, fp, fn, tn)

        return None

    def _check_f1_compatibility(
        self, known_metrics: typing.Mapping[str, float]
    ) -> ValidationResult:
        """检查F1分数与precision/recall的兼容性"""

        f1 = known_metrics['f1_score']

        if 'precision' in known_metrics and 'recall' in known_metrics:
            precision = known_metrics['precision']
            recall = known_metrics['recall']

            # 计算理论F1值
            if precision + recall == 0:
                expected_f1 = 0
            else:
                expected_f1 = 2 * precision * recall / (precision + recall)

            if abs(f1 - expected_f1) > 0.01:  # 1%的容忍度
                return ValidationResult(
                    False,
                    f'F1分数({f1})与precision({precision})和recall({recall})不兼容，'
                    f'期望F1值为{expected_f1:.3f}',
                )

        return ValidationResult(True)

    def _infer_binary_confusion_matrix(
        self, n_pos: int, n_neg: int, known_metrics: typing.Mapping[str, float]
    ) -> Tuple[np.ndarray, Dict[str, float]]:
        """推断二分类混淆矩阵"""

        # 如果有足够的约束，直接求解
        if self._has_sufficient_constraints(known_metrics):
            return self._solve_confusion_matrix_directly(n_pos, n_neg, known_metrics)

        # 否则使用优化方法或随机生成
        return self._generate_feasible_confusion_matrix(n_pos, n_neg, known_metrics)

    def _generate_feasible_confusion_matrix(
        self, n_pos: int, n_neg: int, known_metrics: typing.Mapping[str, float]
    ) -> Tuple[np.ndarray, Dict[str, float]]:
        """生成一个满足当前约束的二分类混淆矩阵近似解。"""
        return self._solve_confusion_matrix_directly(n_pos, n_neg, known_metrics)

    def _has_sufficient_constraints(
        self, known_metrics: typing.Mapping[str, float]
    ) -> bool:
        """检查是否有足够的约束来确定混淆矩阵"""

        # 需要至少2个独立的约束
        sufficient_combinations = [
            {'accuracy', 'precision'},
            {'accuracy', 'recall'},
            {'accuracy', 'specificity'},
            {'precision', 'recall'},
            {'precision', 'specificity'},
            {'recall', 'specificity'},
        ]

        known_set = set(known_metrics.keys())
        return any(combo.issubset(known_set) for combo in sufficient_combinations)

    def _solve_confusion_matrix_directly(
        self, n_pos: int, n_neg: int, known_metrics: typing.Mapping[str, float]
    ) -> Tuple[np.ndarray, Dict[str, float]]:
        """直接求解混淆矩阵"""

        if 'precision' in known_metrics and 'recall' in known_metrics:
            # 原有的 precision + recall 组合
            precision = known_metrics['precision']
            recall = known_metrics['recall']

            # 从recall计算TP
            tp = recall * n_pos
            fn = n_pos - tp

            # 从precision计算FP
            if precision == 0:
                if tp > 0:
                    raise ValueError('precision=0但recall>0，数学上不可能')
                fp = 0
            else:
                fp = tp / precision - tp

            tn = n_neg - fp

        elif 'accuracy' in known_metrics and 'precision' in known_metrics:
            # accuracy + precision 组合
            accuracy = known_metrics['accuracy']
            precision = known_metrics['precision']

            n_samples = n_pos + n_neg

            if precision == 0:
                # precision=0 意味着 TP=0
                tp = 0
                fp = 0
                fn = n_pos
                tn = accuracy * n_samples
            elif abs(precision - 0.5) < 1e-9:
                # precision ≈ 0.5 的特殊情况
                if abs(n_neg - accuracy * n_samples) > 1e-9:
                    raise ValueError(
                        f'指标组合不可行: 当 precision=0.5 时，'
                        f'必须满足 n_neg ({n_neg}) ≈ accuracy * n_samples ({accuracy * n_samples:.2f})，'
                        f'但实际差异为 {abs(n_neg - accuracy * n_samples):.2f}'
                    )
                # 欠约束情况，生成一个合理的解
                warnings.warn(
                    '指标组合欠约束: precision=0.5 且约束条件满足时，'
                    '存在多个解，已生成一个可能的解（假设 recall=0.5）'
                )
                tp = n_pos * 0.5
                fn = n_pos - tp
                tn = accuracy * n_samples - tp
                fp = n_neg - tn
            else:
                # 标准情况：求解线性方程组
                denominator = 1 - 2 * precision
                tp = precision * (n_neg - accuracy * n_samples) / denominator
                tn = accuracy * n_samples - tp
                fp = n_neg - tn
                fn = n_pos - tp

        elif 'accuracy' in known_metrics and 'recall' in known_metrics:
            # accuracy + recall 组合
            accuracy = known_metrics['accuracy']
            recall = known_metrics['recall']

            n_samples = n_pos + n_neg

            # TP = recall * n_pos
            tp = recall * n_pos
            fn = n_pos - tp

            # TN = accuracy * n_samples - TP
            tn = accuracy * n_samples - tp
            fp = n_neg - tn

        elif 'accuracy' in known_metrics and 'specificity' in known_metrics:
            # accuracy + specificity 组合
            accuracy = known_metrics['accuracy']
            specificity = known_metrics['specificity']

            n_samples = n_pos + n_neg

            # TN = specificity * n_neg
            tn = specificity * n_neg
            fp = n_neg - tn

            # TP = accuracy * n_samples - TN
            tp = accuracy * n_samples - tn
            fn = n_pos - tp

        elif 'precision' in known_metrics and 'specificity' in known_metrics:
            # precision + specificity 组合
            precision = known_metrics['precision']
            specificity = known_metrics['specificity']

            # TN = specificity * n_neg
            tn = specificity * n_neg
            fp = n_neg - tn

            if precision == 0:
                tp = 0
            elif abs(precision - 1.0) < 1e-9:
                # precision ≈ 1.0 的特殊情况
                if abs(fp) > 1e-9:
                    raise ValueError(
                        f'指标组合不可行: precision=1 要求 FP=0，'
                        f'但根据 specificity 推断 FP={fp:.2f}'
                    )
                # 欠约束情况，生成一个合理的解
                warnings.warn(
                    '指标组合欠约束: precision=1 且 FP=0 时，'
                    '存在多个解，已生成一个可能的解（假设 recall=1）'
                )
                tp = n_pos  # 假设完美分类器
                fn = 0
            else:
                # 标准情况
                tp = precision * fp / (1 - precision)
                fn = n_pos - tp

        elif 'recall' in known_metrics and 'specificity' in known_metrics:
            # recall + specificity 组合
            recall = known_metrics['recall']
            specificity = known_metrics['specificity']

            # TP = recall * n_pos
            tp = recall * n_pos
            fn = n_pos - tp

            # TN = specificity * n_neg
            tn = specificity * n_neg
            fp = n_neg - tn

        else:
            # 其他组合暂未实现
            available_combinations = [
                'precision + recall',
                'accuracy + precision',
                'accuracy + recall',
                'accuracy + specificity',
                'precision + specificity',
                'recall + specificity',
            ]
            current_metrics = list(known_metrics.keys())
            raise NotImplementedError(
                f'约束组合 {current_metrics} 暂未实现。'
                f'支持的组合包括: {available_combinations}'
            )

        # 验证推导出的浮点值是否在理论范围内
        self._validate_derived_values(tp, fp, fn, tn, n_pos, n_neg)

        # 舍入到整数并确保非负
        tp, fp, fn, tn = [max(0, round(x)) for x in [tp, fp, fn, tn]]

        # 调整以严格满足样本数约束
        if tp + fn != n_pos:
            fn = n_pos - tp
            if fn < 0:
                fn = 0
                tp = n_pos

        if fp + tn != n_neg:
            tn = n_neg - fp
            if tn < 0:
                tn = 0
                fp = n_neg

        # 验证结果的合理性
        if tp < 0 or fp < 0 or fn < 0 or tn < 0:
            raise ValueError(f'求解得到负值: TP={tp}, FP={fp}, FN={fn}, TN={tn}')

        confusion_matrix = np.array([[tn, fp], [fn, tp]], dtype=int)
        inferred_metrics = self._calculate_all_metrics_binary(confusion_matrix)

        return confusion_matrix, inferred_metrics

    def _validate_derived_values(
        self, tp: float, fp: float, fn: float, tn: float, n_pos: int, n_neg: int
    ) -> None:
        """验证推导出的混淆矩阵元素是否在合理范围内

        Parameters
        ----------
        - `tp` : `float`
            - 真正例数量
        - `fp` : `float`
            - 假正例数量
        - `fn` : `float`
            - 假负例数量
        - `tn` : `float`
            - 真负例数量
        - `n_pos` : `int`
            - 正样本总数
        - `n_neg` : `int`
            - 负样本总数

        Raises
        ------
        ValueError
            当推导出的值超出合理范围时
        """
        # 允许小的负容忍度以应对浮点误差
        tolerance = 1e-9

        # 检查每个值是否在合理范围内
        if (
            tp > n_pos + tolerance
            or tp < -tolerance
            or fn > n_pos + tolerance
            or fn < -tolerance
            or fp > n_neg + tolerance
            or fp < -tolerance
            or tn > n_neg + tolerance
            or tn < -tolerance
        ):
            raise ValueError(
                f'指标组合不可行: 推导出的混淆矩阵元素超出范围 '
                f'(TP={tp:.2f}, FP={fp:.2f}, FN={fn:.2f}, TN={tn:.2f})。'
                f'有效范围: TP,FN ∈ [0,{n_pos}], FP,TN ∈ [0,{n_neg}]'
            )

        # 检查行列和是否正确
        if abs((tp + fn) - n_pos) > tolerance:
            raise ValueError(
                f'指标组合不可行: 正样本分布不一致 '
                f'(TP + FN = {tp + fn:.2f}, 期望 = {n_pos})'
            )

        if abs((fp + tn) - n_neg) > tolerance:
            raise ValueError(
                f'指标组合不可行: 负样本分布不一致 '
                f'(FP + TN = {fp + tn:.2f}, 期望 = {n_neg})'
            )

    def _calculate_all_metrics_binary(
        self, confusion_matrix: np.ndarray
    ) -> Dict[str, float]:
        """计算所有二分类指标"""

        tn, fp, fn, tp = confusion_matrix.ravel()

        metrics = {}
        total = tp + tn + fp + fn

        metrics['accuracy'] = (tp + tn) / total if total > 0 else 0
        metrics['precision'] = tp / (tp + fp) if (tp + fp) > 0 else 0
        metrics['recall'] = tp / (tp + fn) if (tp + fn) > 0 else 0
        metrics['specificity'] = tn / (tn + fp) if (tn + fp) > 0 else 0

        if metrics['precision'] + metrics['recall'] > 0:
            metrics['f1_score'] = (
                2
                * metrics['precision']
                * metrics['recall']
                / (metrics['precision'] + metrics['recall'])
            )
        else:
            metrics['f1_score'] = 0

        return metrics

    def _validate_inferred_metrics(
        self,
        known_metrics: typing.Mapping[str, float],
        inferred_metrics: typing.Mapping[str, float],
    ) -> ValidationResult:
        """验证推断结果是否与已知指标一致"""

        tolerance = 0.05  # 5%的容忍度

        for metric_name, expected_value in known_metrics.items():
            if metric_name in inferred_metrics:
                actual_value = inferred_metrics[metric_name]
                if abs(actual_value - expected_value) > tolerance:
                    return ValidationResult(
                        False,
                        f'推断的{metric_name}({actual_value:.3f})与期望值({expected_value:.3f})差异过大',
                    )

        return ValidationResult(True)

    def _validate_multiclass_classification(
        self, class_distribution: typing.Sequence[float]
    ) -> ValidationResult:
        """验证多分类问题（简化实现）"""

        # 计算各类别样本数
        class_counts = [
            round(self.spec.n_samples * ratio) for ratio in class_distribution
        ]

        # 调整确保总和正确
        diff = self.spec.n_samples - sum(class_counts)
        class_counts[0] += diff

        try:
            confusion_matrix = self._generate_feasible_multiclass_confusion_matrix(
                class_counts
            )
            inferred_metrics = self._calculate_all_metrics_multiclass(confusion_matrix)

            return ValidationResult(
                True, inferred_metrics=inferred_metrics, confusion_matrix=confusion_matrix
            )
        except Exception as e:
            return ValidationResult(False, f'多分类指标组合不可行: {str(e)}')

    def _generate_feasible_multiclass_confusion_matrix(
        self, class_counts: typing.Sequence[int]
    ) -> np.ndarray:
        """生成可行的多分类混淆矩阵"""

        n_classes = len(class_counts)
        confusion_matrix = np.zeros((n_classes, n_classes), dtype=int)

        target_accuracy = self.spec.accuracy if self.spec.accuracy is not None else 0.8

        for i in range(n_classes):
            class_size = class_counts[i]
            if class_size == 0:
                continue

            # 正确分类的数量
            correct_count = max(0, min(class_size, round(class_size * target_accuracy)))
            incorrect_count = class_size - correct_count

            # 分配正确分类
            confusion_matrix[i, i] = correct_count

            # 分配错误分类
            if incorrect_count > 0:
                other_classes = [j for j in range(n_classes) if j != i]
                if other_classes:
                    errors = self.rng.multinomial(
                        incorrect_count, [1 / len(other_classes)] * len(other_classes)
                    )
                    for j, error_count in enumerate(errors):
                        confusion_matrix[i, other_classes[j]] = error_count

        return confusion_matrix

    def _calculate_all_metrics_multiclass(
        self, confusion_matrix: np.ndarray
    ) -> Dict[str, float]:
        """计算多分类指标"""

        n_classes = confusion_matrix.shape[0]
        metrics = {}

        # 总体准确率
        metrics['accuracy'] = np.trace(confusion_matrix) / np.sum(confusion_matrix)

        # 每类指标
        precisions = []
        recalls = []
        f1_scores = []

        for i in range(n_classes):
            tp = confusion_matrix[i, i]
            fp = np.sum(confusion_matrix[:, i]) - tp
            fn = np.sum(confusion_matrix[i, :]) - tp

            precision = tp / (tp + fp) if (tp + fp) > 0 else 0
            recall = tp / (tp + fn) if (tp + fn) > 0 else 0
            f1 = (
                2 * precision * recall / (precision + recall)
                if (precision + recall) > 0
                else 0
            )

            precisions.append(precision)
            recalls.append(recall)
            f1_scores.append(f1)

        metrics['precision_macro'] = np.mean(precisions)
        metrics['recall_macro'] = np.mean(recalls)
        metrics['f1_score_macro'] = np.mean(f1_scores)

        return metrics


def validate_classification_metrics(spec: ClassificationSpec) -> ValidationResult:
    """验证分类指标并推断缺失值"""
    validator = ClassificationMetricsValidator(spec)
    return validator.validate_and_infer()


# 测试示例
if __name__ == '__main__':
    # 示例1: 二分类，给定accuracy和precision
    print('=== 示例1: 可行的指标组合 ===')
    spec1 = ClassificationSpec(
        n_samples=1000,
        n_classes=2,
        class_distribution=[0.6, 0.4],
        accuracy=0.85,
        precision=0.80,
        random_seed=42,
    )

    result1 = validate_classification_metrics(spec1)
    print(f'可行性: {result1.is_feasible}')
    if result1.is_feasible:
        print(f'推断指标: {result1.inferred_metrics}')
        print(f'混淆矩阵:\n{result1.confusion_matrix}')
    else:
        print(f'错误: {result1.error_message}')

    # 示例2: 不可行的指标组合
    print('\n=== 示例2: 不可行的指标组合 ===')
    spec2 = ClassificationSpec(
        n_samples=100,
        n_classes=2,
        accuracy=0.95,
        precision=0.50,
        recall=0.95,
        random_seed=42,
    )

    result2 = validate_classification_metrics(spec2)
    print(f'可行性: {result2.is_feasible}')
    if not result2.is_feasible:
        print(f'错误: {result2.error_message}')
    else:
        print(f'推断指标: {result2.inferred_metrics}')
        print(f'混淆矩阵:\n{result2.confusion_matrix}')

    # 示例3: 另一个不可行的组合
    print('\n=== 示例3: 另一个不可行的组合 ===')
    spec3 = ClassificationSpec(
        n_samples=50,
        n_classes=2,
        accuracy=0.99,
        precision=0.10,
        recall=0.99,
        random_seed=42,
    )

    result3 = validate_classification_metrics(spec3)
    print(f'可行性: {result3.is_feasible}')
    if not result3.is_feasible:
        print(f'错误: {result3.error_message}')
    else:
        print(f'推断指标: {result3.inferred_metrics}')



from collections import OrderedDict
from typing import Dict, Union

import numpy as np
import torch
from torch import nn


def module_grad_norm(module: nn.Module, p_norm: int = 2) -> float:
    """计算 module 当前所有参数的梯度的范数

    moduleGradNorm 的计算结果为 module 中所有子模块的 parameter 组成的向量的 norm

    Parameters
    ----------
    - `module` : `torch.nn.Module`
        - 要计算梯度范数的模块
    - `p_norm` : `int`, optional
        - 范数阶数, by default 2

    Returns
    ----------
    - `float`
        - 所有参数梯度的总范数
    """
    sum_grad_norm = 0.0
    for p in module.parameters():
        if p.grad is None:
            continue
        grad = p.grad.detach().norm(p_norm).item()
        sum_grad_norm += grad**p_norm
    total_grad_norm = sum_grad_norm ** (1 / p_norm)
    return total_grad_norm


def param_grad_norm(module: nn.Module, p_norm: int = 2) -> Dict[str, float]:
    """分别计算 module 内各个 parameter 梯度的范数

    以字典的格式返回每个参数的梯度范数。

    Parameters
    ----------
    - `module` : `torch.nn.Module`
        - 要计算梯度范数的模块
    - `p_norm` : `int`, optional
        - 范数阶数, by default 2

    Returns
    ----------
    - `Dict[str, float]`
        - 参数名到其梯度范数的映射字典，格式为 `{parameter_name: norm}`
    """
    norms = {}
    for name, p in module.named_parameters():
        if p.grad is not None:
            grad = p.grad.detach().norm(p_norm).item()
            norms[name] = grad
    return norms


def grad_norm(
    module: nn.Module, p_norm: int = 2, type: str = 'module'
) -> Union[float, dict[str, float]]:
    """计算 module 梯度的范数

    根据 type 的不同，采用不同的方式计算梯度范数。本质上是调用其他两个函数。

    Parameters
    ----------
    - `module` : `torch.nn.Module`
        - 要计算梯度范数的模块
    - `p_norm` : `int`, optional
        - 范数阶数, by default 2
    - `type` : `str`, optional
        - 返回类型, by default 'module'
        - 'module': 计算整个 module 梯度的范数，调用 `moduleGradNorm`，返回 float
        - 'param': 计算 module 各个 parameter 的范数，调用 `paramGradNorm`，返回 dict

    Returns
    ----------
    - `Union[float, dict]`
        - 当 type='module' 时返回 float，表示总范数
        - 当 type='param' 时返回 dict，表示各参数范数

    Raises
    ----------
    - `ValueError`
        - 当 type 参数不是 'module' 或 'param' 时抛出
    """
    if type == 'module':
        return module_grad_norm(module, p_norm)
    elif type == 'param':
        return param_grad_norm(module, p_norm)
    else:
        raise ValueError(f'Invalid argument type: {type}')



from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Generic, Mapping, Sequence

from simple_inject import provide as inject_provide

from zyplib.experiment._types import (
    ParamsT,
    RecordConfig,
    TrialMode,
    matches_export_rule,
    serialize_value,
)
from zyplib.utils.fs import write_yaml


def _merge_records(
    store: dict[str, dict[str, Any]], namespace: str, key: str, value: Any
) -> None:
    store.setdefault(namespace, {})[key] = value


def _clone_records(records: Mapping[str, Mapping[str, Any]]) -> dict[str, dict[str, Any]]:
    return {namespace: dict(values) for namespace, values in records.items()}


@dataclass(slots=True)
class ExperimentContext(Generic[ParamsT]):
    """run 级上下文。

    `prepare_run()` 会在这个对象上工作，用来准备跨 trial 共享的数据，
    例如文件清单、固定 fold plan、预扫描索引等。

    Parameters
    ----------
    - `name`: `str`, experiment 名称
    - `params`: `ParamsT`, 当前运行参数
    - `run_dir`: `Path`, 当前 run 的输出目录
    - `record`: `RecordConfig | None`, record 行为配置
    - `shared`: `Any`, run 级共享对象, by default `None`

    Examples
    ----------
    ```python
    def prepare_run(ctx):
        ctx.set('subject_count', 12, namespace='result')
        return {'all_files': files}
    ```
    """

    name: str
    params: ParamsT
    run_dir: Path
    record: RecordConfig | None = None
    shared: Any = None
    _records: dict[str, dict[str, Any]] = field(
        default_factory=dict, init=False, repr=False
    )

    def set(self, key: str, value: Any, namespace: str = 'default') -> None:
        """写入 run 级 namespace 状态。"""

        _merge_records(self._records, namespace, key, value)

    def recorded_states(self) -> dict[str, dict[str, Any]]:
        """返回 run 级已记录状态。"""

        return _clone_records(self._records)

    def write_yaml(self, relative_path: str, data: Mapping[str, Any]) -> Path:
        """把 run 级数据写到当前 experiment 目录下。"""

        path = self.run_dir / relative_path
        write_yaml(path, serialize_value(data))
        return path


@dataclass(slots=True)
class TrialContext(Generic[ParamsT]):
    """trial 级上下文。

    trial 的数据准备、模型构建和训练逻辑通常都写在这里的消费链条上。
    它暴露了参数、trial 元信息、shared 数据与记录接口。

    Parameters
    ----------
    - `experiment`: `ExperimentContext[ParamsT]`, 所属 run 的上下文
    - `trial_name`: `str`, 当前 trial 名称
    - `trial_index`: `int`, 0-based trial 索引
    - `trial_mode`: `TrialMode`, 当前 trial 模式
    - `trial_seed`: `int | None`, 当前 trial 使用的随机种子
    - `trial_dir`: `Path`, 当前 trial 的输出目录

    Examples
    ----------
    ```python
    def build_data(ctx):
        print(ctx.trial_name, ctx.trial_seed)
        return data
    ```
    """

    experiment: ExperimentContext[ParamsT]
    trial_name: str
    trial_index: int
    trial_mode: TrialMode
    trial_seed: int | None
    trial_dir: Path
    _records: dict[str, dict[str, Any]] = field(
        default_factory=dict, init=False, repr=False
    )

    @property
    def params(self) -> ParamsT:
        """返回当前 trial 共享的参数对象。"""

        return self.experiment.params

    @property
    def run_dir(self) -> Path:
        """返回所属 experiment 的 run 目录。"""

        return self.experiment.run_dir

    @property
    def shared(self) -> Any:
        """返回 `prepare_run()` 生成的 run 级 shared 数据。"""

        return self.experiment.shared

    @property
    def record(self) -> RecordConfig | None:
        """返回当前 experiment 的 record 配置。"""

        return self.experiment.record

    @property
    def trial_no(self) -> int:
        """返回从 1 开始计数的 trial 编号。"""

        return self.trial_index + 1

    def set(self, key: str, value: Any, namespace: str = 'default') -> None:
        """写入 trial 级 namespace 状态，并按配置决定是否注入当前 scope。"""

        _merge_records(self._records, namespace, key, value)
        if (
            self.record
            and self.record.enabled
            and namespace in self.record.inject_namespaces
        ):
            inject_provide(key, value, namespace=namespace)

    def recorded_states(
        self, namespaces: Sequence[str] | None = None
    ) -> dict[str, dict[str, Any]]:
        """返回 trial 本地已记录状态。"""

        records = (
            self._records
            if namespaces is None
            else {namespace: self._records.get(namespace, {}) for namespace in namespaces}
        )
        return _clone_records(records)

    def snapshot_states(
        self, namespaces: Sequence[str] | None = None
    ) -> dict[str, dict[str, Any]]:
        """按 namespace 抓取当前 trial 的状态快照。"""

        if namespaces is None and self.record and self.record.enabled:
            filtered: dict[str, dict[str, Any]] = {}
            for namespace, values in self._records.items():
                exported = {
                    key: value
                    for key, value in values.items()
                    if matches_export_rule(self.record.export_rule, namespace, key, value)
                }
                if exported:
                    filtered[namespace] = exported
            return _clone_records(filtered)
        return self.recorded_states(namespaces)

    def write_yaml(self, relative_path: str, data: Mapping[str, Any]) -> Path:
        """把 trial 级数据写到当前 trial 目录下。"""

        path = self.trial_dir / relative_path
        write_yaml(path, serialize_value(data))
        return path


__all__ = ['ExperimentContext', 'TrialContext']



import typing
import fnmatch
from copy import deepcopy
from typing import Callable, Dict, List, Literal, Optional

# import matplotlib.pyplot as plt
import torch
from torchmetrics import Metric, MetricCollection

from zyplib.train.utils import use_device
from zyplib.utils.ensure import ensure_scalar
from zyplib.utils.print import print_debug, print_warning, warning
from zyplib.utils.kv import prefixed_dict

ReducerName = Literal['mean', 'sum']
ActivationFn = Callable[[torch.Tensor], torch.Tensor]


class ScalarMetric:
    def __init__(
        self,
        default_reducer: ReducerName = 'mean',
        retain_batch_history: bool = False,
        retain_epoch_history: bool = True,
        auto_define: bool = False,
    ):
        """记录单个 float 类型的指标

        唯一的作用是简化「记录 Batch 指标 - 在 epoch 结束后求平均」的工作流程

        Parameters
        ----------
        - `default_reducer` : `Literal['mean', 'sum']`, optional
            - 默认的聚合方式, by default 'mean'
        - `retain_batch_history` : `bool`, optional
            - 是否保留每个 batch 的历史记录, by default False
        - `retain_epoch_history` : `bool`, optional
            - 是否保留每个 epoch 的历史记录, by default True
        - `auto_define` : `bool`, optional
            - 是否自动定义未知的指标，当为 True 时，update() 中传入的未定义指标会自动使用 default_reducer 进行定义, by default False

        Example
        --------

        ```py
        # 手动模式：需要预先定义指标
        metric = ScalarMetric()
        metric.define_metric('loss', reducer='mean')
        metric.define_metric('acc', reducer='mean')

        # 自动定义模式：无需预先定义
        metric = ScalarMetric(auto_define=True)

        for batch in dataloader:
            y_hat, y = batch
            loss = compute_loss(y_hat, y)
            acc = compute_acc(y_hat, y)
            metric.update(loss=loss.item(), acc=acc)  # 自动定义 loss 和 acc
        metric.compute()  # 得到当前 epoch 的平均损失和准确率
        ```
        """
        self._batch_metrics: dict[
            str, list[float]
        ] = {}  # 每个 batch 内的指标, 随着 reset 清空

        self._batch_history: list[dict[str, float]] = []
        self._epoch_history: list[dict[str, float]] = []

        self._metrics: dict[str, Callable[[list[float]], float]] = {}
        self._default_reducer = (
            lambda x: sum(x) / len(x) if default_reducer == 'mean' else sum(x)
        )
        self._default_reducer_name: ReducerName = default_reducer

        self._retain_batch_history = retain_batch_history
        self._retain_epoch_history = retain_epoch_history
        self._auto_define = auto_define

        self._undefined_metric_name = set()

    def clone(self):
        """克隆一个完全相同的对象"""
        return deepcopy(self)

    def define_metric(self, metric_name: str, reducer: ReducerName = 'mean'):
        """定义一个新的指标及其计算方式

        Parameters
        ----------
        - `metric_name` : `str`
            - 指标名称
        - `reducer` : `Literal['mean', 'sum']`, optional
            - 计算方式, 默认为 'mean'
        """
        self._batch_metrics[metric_name] = []
        if reducer == 'mean':
            self._metrics[metric_name] = lambda x: sum(x) / len(x)
        elif reducer == 'sum':
            self._metrics[metric_name] = sum
        else:
            raise ValueError("reducer must be either 'mean' or 'sum'")

    def update(self, **metrics: float):
        """更新当前 step 的指标值

        Parameters
        ----------
        **metrics : float
            指标名称和对应的值
        """
        for name, value in metrics.items():
            if name not in self._metrics:
                if self._auto_define:
                    # 自动定义模式：使用默认 reducer 自动定义指标
                    self.define_metric(name, self._default_reducer_name)
                else:
                    # 传统模式：跳过未定义的指标并警告
                    if name not in self._undefined_metric_name:
                        print_warning(f'Metric {name} not defined, skipping')
                        self._undefined_metric_name.add(name)
                    continue
            if isinstance(value, torch.Tensor):
                # 以防万一
                value = value.item()
            self._batch_metrics[name].append(value)

        if self._retain_batch_history:
            self._batch_history.append(metrics)

    def compute(self, auto_reset: bool = False) -> dict[str, float]:
        """reduce 当前 batch 的指标

        建议在每个 epoch 结束之后调用, 并且调用 compute 之后需要手动调用 reset 清空当前 step 的数据

        Returns
        -------
        dict[str, float]
            指标名称和对应的计算结果
        """
        results = {}
        for name, reducer in self._metrics.items():
            if self._batch_metrics[name]:  # 确保有数据
                results[name] = reducer(self._batch_metrics[name])

        # 保存结果并清空当前 step 的数据
        if self._retain_epoch_history:
            self._epoch_history.append(results)

        if auto_reset:
            self.reset()

        return results

    def reset(self):
        self._batch_metrics = {name: [] for name in self._metrics.keys()}


class TorchMetricRecorder:
    def __init__(
        self,
        metrics: MetricCollection | typing.Mapping[str, Metric] | None = None,
        before_metric: Literal['sigmoid', 'softmax']
        | ActivationFn
        | typing.Mapping[str | Literal['*'], ActivationFn]
        | None = None,
        scalar_metrics: ScalarMetric
        | typing.Sequence[str]
        | str
        | Literal['auto']
        | None = 'loss',
    ):
        """添加基于 `torchmetrics` 的日志评估指标的功能

        Trainstep-wise Metric --> 记录 Metric Value --> Epoch --> 计算 Epochwise Metric (Reduce) --> 记录 Epochwise Metric

        区分两种类型:

        - 需要计算的指标，如 Accuracy 等；这些传入 torchmetric.Metric 或者 MetricCollection 对象
        - 需要记录的标量数据，如 loss 等; 传入名称字符串
            - 这些标量指标来源，应当是各个 training_step 返回的字典中包含的变量
            - 例如默认内置的 training_step 返回的字典中为 `{'loss': loss}`，所以 TorchMetricRecorder 默认会记录 loss 指标

        Parameters
        ----------
        - `metrics` : `MetricCollection | Dict[str, Metric]`, optional
            - 评估指标; 指哪些需要通过计算得到的指标，例如 Accuracy 等; 依赖于 torchmetrics 库
            - 使用 `torchmetrics.MetricCollection` 或 `dict` 类型
        - `before_metric` : `'sigmoid' | 'softmax' | Callable | dict[str, Callable]`, optional
            - 在计算评估指标之前的激活函数: `(y_hat) -> y_activated`
            - 例如在二分类问题中，模型最后一层没有激活函数, 在计算评估指标 (即 `metrics`) 之前需要使用 `torch.sigmoid` 激活
        - `scalar_metrics` : `ScalarMetric | str | list[str] | Literal['auto']`, optional
            - 用于记录一些外部传进来的标量数据，例如 loss 等
            - 如果传入 `ScalarMetric`，则直接使用
                - 通过 `self.update` 方法中的 `**metrics` 参数传入的指标都会被 update 到 `scalar_metrics` 中
            - 如果传入 `str`，则自动创建一个 ScalarMetric 并定义该指标 (mean)
            - 如果传入 `list[str]`, 则自动创建一个 ScalarMetric 并定义这些指标 (mean)
            - 如果传入 `'auto'`，则创建一个启用 auto_define 的 ScalarMetric，自动记录所有传入的指标

        `before_metric`:
        ----------
        在计算评估指标之前的激活函数: `(y_hat) -> y_activated`

        例如在二分类问题中，模型最后一层没有激活函数, 在计算评估指标 (即 `metrics`) 之前需要使用 `torch.sigmoid` 激活

        - `None`: 不需要额外的激活函数
        - `Literal['sigmoid', 'softmax']`: 使用 torch.sigmoid 和 torch.softmax
        - `(Tensor) -> Tensor`: 指定的激活函数
        - `dict[str, (Tensor) -> Tensor]`: 在比较复杂的情况下, 各个指标使用的输入不同的情况下可以用这个
            - key: 支持 glob 风格匹配模式：
                - 精确匹配：`"accuracy"` 匹配名为 "accuracy" 的指标
                - 通配符：`"*"` 匹配所有指标
                - glob 模式：`"U*"` 匹配所有以 "U" 开头的指标，`"train_*"` 匹配所有以 "train_" 开头的指标
            - value: `(Tensor) -> Tensor`, 为 None，则不激活
        """
        if before_metric == 'sigmoid':
            self.before_metric = torch.sigmoid
        elif before_metric == 'softmax':
            self.before_metric = torch.softmax
        else:
            self.before_metric = before_metric

        # 只保留一个 metrics collection
        if metrics is None:
            self.metrics = MetricCollection({})
        elif isinstance(metrics, dict):
            self.metrics = MetricCollection(
                typing.cast(dict[str, Metric | MetricCollection], metrics)
            )
        else:
            self.metrics = metrics

        # 记录哪些传入 update 的标量数据, 包括 loss 等
        if isinstance(scalar_metrics, ScalarMetric):
            self.scalar_metrics = scalar_metrics
        elif scalar_metrics == 'auto':
            # 自动模式：创建启用 auto_define 的 ScalarMetric
            self.scalar_metrics = ScalarMetric(default_reducer='mean', auto_define=True)
        else:
            self.scalar_metrics = ScalarMetric(default_reducer='mean')
            if isinstance(scalar_metrics, str):
                self.scalar_metrics.define_metric(scalar_metrics, 'mean')
            elif isinstance(scalar_metrics, (list, tuple)):
                for metric_name in scalar_metrics:
                    self.scalar_metrics.define_metric(metric_name, 'mean')
            elif scalar_metrics is not None:
                warning(f'不符合标准的参数 scalar_metrics: {scalar_metrics}')
            # else:
            #     self.scalar_metrics.define_metric('loss', 'mean')

        # self.step_history = {metric_name: [] for metric_name in self.metrics.keys()}
        # self.step_history['loss'] = []
        # self.epoch_history: list[dict[str, float | torch.Tensor]] = []

        self.current_epoch = 0
        self.current_step = 0

        # 激活函数匹配缓存
        self._activation_cache: Dict[str, ActivationFn | None] = {}

    def _find_matching_activation(self, metric_name: str) -> ActivationFn | None:
        """为指定的指标名称找到匹配的激活函数

        匹配优先级：精确匹配 > glob 模式匹配
        使用缓存机制提升性能，只在第一次匹配时进行计算

        Parameters
        ----------
        metric_name : str
            指标名称

        Returns
        -------
        Callable | None
            匹配的激活函数，如果没有匹配则返回 None
        """
        # 检查缓存
        if metric_name in self._activation_cache:
            return self._activation_cache[metric_name]

        if not isinstance(self.before_metric, dict):
            self._activation_cache[metric_name] = None
            return None

        activation_fn = None
        matched_pattern = None

        # 1. 精确匹配
        if metric_name in self.before_metric:
            activation_fn = self.before_metric[metric_name]
            matched_pattern = f"'{metric_name}' (精确匹配)"
        else:
            # 2. glob 模式匹配
            for pattern, fn in self.before_metric.items():
                if pattern != metric_name and fnmatch.fnmatch(metric_name, pattern):
                    activation_fn = fn
                    matched_pattern = f"'{pattern}' (glob匹配)"
                    break

        # 缓存结果并输出调试信息
        self._activation_cache[metric_name] = activation_fn

        if matched_pattern:
            fn_name = getattr(activation_fn, '__name__', str(activation_fn))
            print_debug(
                f"Metric '{metric_name}' 匹配处理器 {matched_pattern} -> {fn_name}"
            )
        else:
            print_debug(f"Metric '{metric_name}' 无匹配处理器，使用原始输入")

        return activation_fn

    def to_device(self, device: str | torch.device):
        # 简化 to_device，只需要处理一个 metrics collection
        for key, metric in self.metrics.items():
            use_device(metric, device)

    def _apply_activation(
        self, y_hat: torch.Tensor
    ) -> torch.Tensor | dict[str, torch.Tensor]:
        # 处理 before_metric 激活
        if self.before_metric is None:
            y_hat_activated = y_hat
        elif isinstance(self.before_metric, dict):
            # 如果是字典类型，需要对不同的指标使用不同的激活函数
            y_hat_activated = {}
            for metric_name in self.metrics.keys():
                # 使用 glob 匹配查找激活函数
                metric_name = typing.cast(str, metric_name)
                activation_fn = self._find_matching_activation(metric_name)

                if activation_fn is not None:
                    y_hat_activated[metric_name] = activation_fn(y_hat)
                else:
                    y_hat_activated[metric_name] = y_hat
        else:
            # 如果是单个激活函数
            before_metric = self.before_metric
            if before_metric is torch.softmax:
                y_hat_activated = torch.softmax(y_hat, dim=-1)
            else:
                activation_fn = typing.cast(ActivationFn, before_metric)
                y_hat_activated = activation_fn(y_hat)
        return y_hat_activated

    def update(self, pred, target, **scalar_metrics: float):
        """更新某个 step 的评估指标

        Parameters
        ----------
        - `pred` : `torch.Tensor`
            - 预测值
        - `target` : `torch.Tensor`
            - 真实值
        - `**scalar_metrics` : float
            - 其他标量指标; 默认应该包含 loss
        """
        result = {}

        # 更新其他标量指标
        if scalar_metrics:
            self.scalar_metrics.update(**scalar_metrics)
            result.update(scalar_metrics)

        # 更新 torchmetrics，如果 metrics 不为空
        if self.metrics and len(self.metrics) > 0:
            pred_activated = self._apply_activation(pred)
            for metric_name, metric in self.metrics.items():
                if isinstance(pred_activated, dict):
                    current_y_hat = pred_activated[metric_name]
                else:
                    current_y_hat = pred_activated

                step_metric_value = metric.update(
                    current_y_hat, target
                )  # 记录本 step 的指标  # NOTE 注意这里是否正确

                # 记录计算得到的指标
                self._record_compute_value(result, metric_name, step_metric_value)

        self.current_step += 1
        return result

    def compute(self):
        """计算一个 epoch 的评估指标; 并自动 reset"""
        epoch_metrics = {}

        # 计算标量指标
        scalar_results = self.scalar_metrics.compute()
        epoch_metrics.update(scalar_results)
        self.scalar_metrics.reset()

        # 计算 torchmetrics，如果 metrics 不为空
        if self.metrics and len(self.metrics) > 0:
            for metric_name, metric in self.metrics.items():
                epoch_metric_value = metric.compute()
                # epoch_metrics[metric_name] = epoch_metric_value.item()
                self._record_compute_value(epoch_metrics, metric_name, epoch_metric_value)
                metric.reset()

        self.current_epoch += 1
        self.current_step = 0
        return epoch_metrics

    def reset(self):
        self.scalar_metrics.reset()
        if self.metrics and len(self.metrics) > 0:
            for metric in self.metrics.values():
                metric.reset()

    def _record_compute_value(
        self,
        recorder: dict[str, typing.Any],
        name: str,
        value: float | torch.Tensor | typing.Mapping,
    ):
        """将 metric 计算得到的结果记录在字典当中

        由于 metric 计算得到的结果可能是 dict 类型, 也可能是标量, 所以需要单独处理

        value 如果是 dict, 默认应当是只有一层!
        """
        if isinstance(value, dict):
            recorder.update(prefixed_dict(f'{name}.', value))
        elif isinstance(value, typing.Mapping):
            recorder.update(prefixed_dict(f'{name}.', dict(value)))
        elif not isinstance(value, dict) and value is not None:
            # 能转换成标量就记录标量，不能就记录原值
            scalar = ensure_scalar(value)
            recorder[name] = scalar or value



import numpy as np
from numpy.typing import NDArray


def produce_wave_by_psd(
    f: NDArray, psd: NDArray, sfreq: int, t_second: int, n_sample: int = 1
) -> NDArray:
    """根据功率谱密度生成信号

    传入一个 PSD 的频谱图, 模拟生成一个拥有对应 PSD 的信号

    Parameters
    ----------
    - `f` : `NDArray`
        - 频率轴，单位为Hz
    - `psd` : `NDArray`
        - 输入的功率谱密度图
    - `sfreq` : `int`
        - 采样频率，单位为Hz
    - `t_second` : `int`
        - 信号时长，单位为秒
    - `n_sample` : `int`, optional
        - 生成的信号样本数量, by default 1

    Returns
    ----------
    - `NDArray`
        - 生成的信号，形状为 (n_sample, n_timepoints)
    """
    # 计算信号长度
    n_timepoints = int(sfreq * t_second)

    # 构造频率轴（对应FFT的频率bins）
    freqs = np.fft.fftfreq(n_timepoints, 1.0 / sfreq)

    # 只取正频率部分（包括0频率）
    positive_freqs = freqs[: n_timepoints // 2 + 1]

    # 将输入的PSD插值到FFT频率网格上
    psd_interp = np.interp(positive_freqs, f, psd)

    # 从PSD计算幅度谱 (PSD = |X(f)|^2, 所以 |X(f)| = sqrt(PSD))
    amplitude = np.sqrt(psd_interp)

    # 生成多个信号样本
    signals = np.zeros((n_sample, n_timepoints))

    for i in range(n_sample):
        # 生成随机相位 (0 to 2π)
        phases = np.random.uniform(0, 2 * np.pi, len(positive_freqs))

        # DC分量（0Hz）的相位应该为0（实数）
        phases[0] = 0

        # 如果信号长度为偶数，Nyquist频率的相位也应该为0
        if n_timepoints % 2 == 0:
            phases[-1] = 0

        # 构造复数频谱
        complex_spectrum = amplitude * np.exp(1j * phases)

        # 构造完整的频谱（包括负频率部分）
        # 负频率部分是正频率部分的共轭
        if n_timepoints % 2 == 0:
            # 偶数长度
            full_spectrum = np.concatenate(
                [complex_spectrum, np.conj(complex_spectrum[-2:0:-1])]
            )
        else:
            # 奇数长度
            full_spectrum = np.concatenate(
                [complex_spectrum, np.conj(complex_spectrum[-1:0:-1])]
            )

        # 逆FFT得到时域信号
        signal = np.fft.ifft(full_spectrum).real

        # 归一化信号以匹配PSD的能量
        # 计算信号的实际PSD
        signal_fft = np.fft.fft(signal)
        signal_psd = np.abs(signal_fft[: len(positive_freqs)]) ** 2 / (
            sfreq * n_timepoints
        )

        # 计算缩放因子
        target_power = np.trapz(psd_interp, positive_freqs)
        actual_power = np.trapz(signal_psd, positive_freqs)

        if actual_power > 0:
            scale_factor = np.sqrt(target_power / actual_power)
            signal *= scale_factor

        signals[i] = signal

    return signals if n_sample > 1 else signals[0]


def _simulate_eeg_psd(freqs: np.ndarray) -> np.ndarray:
    """创建类似EEG的功率谱密度 (PSD)

    该函数建模典型静息态EEG信号的PSD特征

    Parameters
    ----------
    - `freqs` : `np.ndarray`
        - 频率数组，单位为Hz

    Returns
    ----------
    - `np.ndarray`
        - EEG样式的功率谱密度
    """
    # 1/f background
    background = 10 / (freqs + 0.1) ** 1.2
    # Delta activity (0.5-4 Hz)
    delta_peak = 20 * np.exp(-(((freqs - 2) / 1) ** 2))
    # Theta activity (4-8 Hz)
    theta_peak = 8 * np.exp(-(((freqs - 6) / 2) ** 2))
    # Alpha peak (8-12 Hz)
    alpha_peak = 15 * np.exp(-(((freqs - 10) / 2) ** 2))
    # Beta activity (13-30 Hz)
    beta_peak = 5 * np.exp(-(((freqs - 20) / 8) ** 2))
    # High frequency rolloff
    rolloff = np.exp(-((freqs / 50) ** 2))

    psd = (background + delta_peak + alpha_peak + beta_peak + theta_peak) * rolloff
    return np.maximum(psd, 0.01)


def produce_fake_eeg_waves(
    duration: int,
    sfreq: int,
    n_signals: int = 1,
    psd_freq_range: tuple[float, float] = (0.2, 60),
    target_psd_resolution: float = 0.25,
) -> NDArray:
    """生成具有典型脑电图（EEG）特征的模拟信号

    Parameters
    ----------
    - `duration` : `int`
        - 信号时长 (秒)
    - `sfreq` : `int`
        - 采样频率 (Hz)
    - `n_signals` : `int`, optional
        - 生成的信号数量, by default 1
    - `psd_freq_range` : `tuple[float, float]`, optional
        - 用于构建目标PSD的频率范围, by default (0.5, 60)
    - `target_psd_resolution` : `int`, optional
        - 用于构建目标PSD的频率分辨率, by default 0.25

    Returns
    ----------
    - `a_signals (NDArray)`: 生成的信号, shape (n_signals, n_timepoints) 当 n_signals > 1 时，或 (n_timepoints,) 当 n_signals = 1 时
    """
    f_min, f_max = psd_freq_range
    num_points = int((f_max - f_min) / target_psd_resolution) + 1
    psd_freq_points = num_points if num_points > 1 else 2
    # 1. 创建目标PSD
    freqs = np.linspace(f_min, f_max, psd_freq_points)
    target_psd = _simulate_eeg_psd(freqs)

    # 2. 使用已有的 `produce_signal_by_psd` 函数生成信号
    a_signals = produce_wave_by_psd(freqs, target_psd, sfreq, duration, n_signals)

    return a_signals



"""EEG connectivity feature extraction algorithms.

Provides various connectivity measures for EEG signals analysis:

- Pearson Correlation
- Phase Locking Value (PLV)
- Phase Lag Index (PLI)
- Coherence
- Weighted Phase Lag Index (wPLI)
- Debiased Squared wPLI
- Cross-correlation
- Coherency

#### References

- [1] https://mne.tools/mne-connectivity/stable/index.html
- [2] Bastos, A. M., & Schoffelen, J. M. (2016). A tutorial review of functional
    connectivity analysis methods and their interpretational pitfalls.
    Frontiers in systems neuroscience, 9, 175.
"""

import typing

from typing import Any, Literal, Optional, Tuple, Union, cast

import numpy as np
from mne_connectivity import spectral_connectivity_epochs
from mne_connectivity.viz import plot_connectivity_circle

from zyplib.utils.ensure import ensure_ndarray

ConnMethod = Literal[
    'coh',  # Coherence
    'cohy',  # Coherency
    'plv',  # Phase Locking Value
    'pli',  # Phase Lag Index
    'wpli',  # Weighted Phase Lag Index
    'wpli2_debiased',  # Debiased Squared wPLI
    'ppc',  # Pairwise Phase Consistency
    'ccd',  # Canonical Correlation Density
]


def compute_spectra_connectivity(
    data: np.ndarray,
    sfreq: float,
    method: ConnMethod = 'pli',
    fmin: Optional[float] = None,
    fmax: Optional[float] = None,
    mode: str = 'multitaper',
    average: bool = True,
    **kwargs,
) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
    """计算 EEG 数据的连接性度量

    本函数封装了 mne-connectivity 以简化使用。

    Parameters
    ----------
    - `data` : `np.ndarray`
        - EEG 数据数组，形状为 [n_trials, n_channels, n_samples] 或 [n_channels, n_samples]
    - `sfreq` : `float`
        - 采样频率，单位为 Hz
    - `method` : `ConnMethod`
        - 计算的连接性度量：
        - 'coh': Coherence
        - 'cohy': Coherency
        - 'plv': Phase Locking Value
        - 'pli': Phase Lag Index
        - 'wpli': Weighted Phase Lag Index
        - 'wpli2_debiased': Debiased Squared wPLI
        - 'ppc': Pairwise Phase Consistency
        - 'ccd': Canonical Correlation Density
    - `fmin` : `float | None`
        - 感兴趣的最小频率。如果为 None，则使用最低可用频率
    - `fmax` : `float | None`
        - 感兴趣的最大频率。如果为 None，则使用最高可用频率
    - `average` : `bool`
        - 是否在频率和试验之间取平均，默认为 True

    Returns
    -------
    - `Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]`
        - 如果 average=True:
            - 连接性矩阵，形状为 [n_channels, n_channels]
        - 如果 average=False:
            - (连接性, 频率) 的元组：
            - 连接性: 形状为 [n_channels, n_channels, n_freqs]
            - 频率: 形状为 [n_freqs,]

    Notes
    -----
    对于基于相位的方法（PLV, PLI），建议进行预处理（例如过滤）。
    """
    data = cast(np.ndarray, ensure_ndarray(data))
    if data.ndim == 2:
        data = data[np.newaxis, ...]  # Add trial dimension
    default_kwargs = {
        'fmin': fmin,
        'fmax': fmax,
        'faverage': average,
        'n_jobs': 1,  # Avoid multiprocessing issues
        'verbose': False,
        'mode': mode,
    }
    kwargs = {**default_kwargs, **kwargs}
    # Compute connectivity using mne-connectivity
    conn = spectral_connectivity_epochs(data, method=method, sfreq=sfreq, **kwargs)

    return conn


def plot_connectivity(
    conn_matrix: np.ndarray,
    node_names: Optional[typing.Sequence] = None,
    title: Optional[str] = None,
    **kwargs,
) -> Any:
    """绘制连接性圆图。

    Parameters
    ----------
    - `conn_matrix` : `np.ndarray`
        - 形状为 (n_channels, n_channels) 的连接性矩阵
    - `node_names` : `list | None`
        - 节点名称列表。如果为 None，则使用数字索引
    - `title` : `str | None`
        - 图表的标题
    - `**kwargs` : `dict`
        - 传递给 mne_connectivity.viz.plot_connectivity_circle 的额外参数

    Notes
    -----
    此函数封装了 mne_connectivity.viz.plot_connectivity_circle，以简化使用。
    """
    if node_names is None:
        node_names = [str(i) for i in range(conn_matrix.shape[0])]

    fig = plot_connectivity_circle(conn_matrix, node_names, title=title, **kwargs)
    return fig


# Convenience functions for common connectivity measures
def coherence(
    data: np.ndarray,
    sfreq: float,
    fmin: Optional[float] = None,
    fmax: Optional[float] = None,
    average: bool = True,
) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
    """Calculate coherence between channels.

    See compute_connectivity for parameter details.
    """
    return compute_spectra_connectivity(data, sfreq, 'coh', fmin, fmax, average=average)


def plv(
    data: np.ndarray,
    sfreq: float,
    fmin: Optional[float] = None,
    fmax: Optional[float] = None,
    average: bool = True,
) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
    """Calculate Phase Locking Value between channels.

    See compute_connectivity for parameter details.
    """
    return compute_spectra_connectivity(data, sfreq, 'plv', fmin, fmax, average=average)


def pli(
    data: np.ndarray,
    sfreq: float,
    fmin: Optional[float] = None,
    fmax: Optional[float] = None,
    average: bool = True,
) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
    """Calculate Phase Lag Index between channels.

    See compute_connectivity for parameter details.
    """
    return compute_spectra_connectivity(data, sfreq, 'pli', fmin, fmax, average=average)


def wpli(
    data: np.ndarray,
    sfreq: float,
    fmin: Optional[float] = None,
    fmax: Optional[float] = None,
    average: bool = True,
) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
    """Calculate Weighted Phase Lag Index between channels.

    See compute_connectivity for parameter details.
    """
    return compute_spectra_connectivity(data, sfreq, 'wpli', fmin, fmax, average=average)



from typing import Literal
from zyplib.nn.modules.base import BaseModule


class ResampleTo(BaseModule):
    """时间序列重采样


    假定输入样本 [N, C, T]

    本模块的作用是将时间维度 T 重采样到指定的长度 out_samples
    """

    def __init__(
        self, out_samples: int, method: Literal['interpolate', 'adaptive_pool'], t_dim=-1
    ): ...



from typing import Any, List, Literal, Optional, Tuple, Union, cast

import numpy as np
import pywt
from scipy.signal import ShortTimeFFT
from scipy.signal.windows import get_window

from zyplib.utils.ensure import ensure_npy

__all__ = ['stft', 'istft', 'spectrogram', 'cwt', 'dwt']


def stft(
    x,
    fs,
    window='hamming',
    nperseg=None,
    noverlap=None,
    nfft=None,
    axis=-1,
    scaling='spectrum',
    return_onesided=True,
    **kwargs,
):
    """Short-Time Fourier Transform (STFT)

    计算短时傅里叶变换，返回复数频谱，所以通常更推荐使用 spectrogram 函数

    Parameters
    ----------
    - `x` : `array_like`
        - 输入信号。支持多维数组，沿指定轴进行变换。
    - `fs` : `float`
        - 采样频率，单位为 Hz。
    - `window` : `str` or `tuple` or `array_like`, optional
        - 窗函数类型或窗函数数组。默认为 'hamming'。
        - 支持 scipy.signal.windows 中的所有窗函数。
    - `nperseg` : `int`, optional
        - 每段的长度。默认为 `int(0.5 * fs)`，即 0.5 秒。
    - `noverlap` : `int`, optional
        - 段间重叠的样本数。默认为 `nperseg // 2`。
    - `nfft` : `int`, optional
        - FFT 的长度。如果大于 `nperseg`，则进行零填充。默认等于 `nperseg`。
    - `axis` : `int`, optional
        - 进行变换的轴。默认为 -1（最后一个轴）。
    - `scaling` : `str`, optional
        - 缩放模式。'spectrum': 功率谱（V²）；'psd': 功率谱密度（V²/Hz）。默认为 'spectrum'。
        - 'spectrum' 返回功率谱，其模的平方是信号的功率谱，单位为 V²。
        - 'psd' 返回功率谱密度，其模的平方是信号的功率谱密度，单位为 V²/Hz。
    - `return_onesided` : `bool`, optional
        - 是否返回单边谱。默认为 True。

    Returns
    ----------
    - `f` : `ndarray`
        - 频率数组，单位为 Hz。
    - `t` : `ndarray`
        - 时间数组，单位为秒。
    - `Zxx` : `ndarray`
        - STFT 结果，复数数组。形状为 (..., freq_bins, time_bins)。

    Examples
    ----------
    ```py
    >>> import numpy as np
    >>> from zyplib.signal.time_frequency import stft
    >>>
    >>> # 生成测试信号
    >>> fs = 1000
    >>> t = np.arange(0, 3, 1/fs)
    >>> x = np.sin(2*np.pi*50*t) + 0.5*np.sin(2*np.pi*120*t)
    >>>
    >>> # 计算 STFT
    >>> f, t_stft, Zxx = stft(x, fs)
    >>> print(f"频率范围: {f[0]:.1f} - {f[-1]:.1f} Hz")
    >>> print(f"时间范围: {t_stft[0]:.2f} - {t_stft[-1]:.2f} s")
    ```
    """
    x = cast(np.ndarray, ensure_npy(x))

    # 设置默认参数
    if nperseg is None:
        nperseg = min(int(0.25 * fs), 1024)  # 更保守的默认值

    if noverlap is None:
        noverlap = nperseg // 2

    if nfft is None:
        nfft = nperseg

    # 计算 hop（步长）
    hop = nperseg - noverlap

    # 获取窗函数
    if isinstance(window, str) or isinstance(window, tuple):
        win = get_window(window, nperseg)
    else:
        win = np.asarray(window)
        if len(win) != nperseg:
            raise ValueError(f'窗长度 ({len(win)}) 必须等于 nperseg ({nperseg})')

    # 确定 FFT 模式
    fft_mode = 'onesided' if return_onesided else 'twosided'

    # 设置缩放模式
    scale_to = 'psd' if scaling == 'psd' else None

    # 创建 ShortTimeFFT 对象
    SFT = ShortTimeFFT(
        win=win, hop=hop, fs=fs, fft_mode=fft_mode, mfft=nfft, scale_to=scale_to, **kwargs
    )

    # 执行 STFT
    Zxx = SFT.stft(x, axis=axis, padding='even')  # 使用 even padding 缓解边缘效应

    # 获取频率和时间数组
    f = SFT.f
    n_samples = x.shape[axis]
    t = SFT.t(n_samples)

    return f, t, Zxx


def istft(
    Zxx,
    fs,
    nperseg,
    window='hamming',
    noverlap=None,
    nfft=None,
    time_axis=-1,
    freq_axis=-2,
    **kwargs,
):
    """Inverse Short-Time Fourier Transform (ISTFT)

    从 STFT 结果重构时域信号。

    Parameters
    ----------
    - `Zxx` : `array_like`
        - STFT 结果的复数数组。
    - `fs` : `float`
        - 采样频率，单位为 Hz。
    - `nperseg` : `int`
        - 每段的长度，必须与计算 STFT 时使用的值相同。
    - `window` : `str` or `tuple` or `array_like`, optional
        - 窗函数类型，应与计算 STFT 时使用的窗函数相同。默认为 'hamming'。
    - `noverlap` : `int`, optional
        - 段间重叠的样本数，应与计算 STFT 时使用的值相同。默认为 `nperseg // 2`。
    - `nfft` : `int`, optional
        - FFT 的长度，应与计算 STFT 时使用的值相同。默认为 `nperseg`。
    - `time_axis` : `int`, optional
        - 时间轴的索引。默认为 -1。
    - `freq_axis` : `int`, optional
        - 频率轴的索引。默认为 -2。

    Returns
    ----------
    - `t` : `ndarray`
        - 重构信号的时间数组。
    - `x` : `ndarray`
        - 重构的时域信号。

    Notes
    -----
    为了确保重构的准确性，必须提供与 STFT 计算时完全相同的参数。
    特别是 `nperseg`、`noverlap` 和 `nfft` 必须与 `stft()` 调用时使用的值一致。
    """
    import warnings

    Zxx = ensure_npy(Zxx)

    # 获取 STFT 矩阵的形状信息
    freq_bins = Zxx.shape[freq_axis]

    # 验证参数
    if noverlap is None:
        noverlap = nperseg // 2
    elif noverlap >= nperseg:
        raise ValueError('noverlap 必须小于 nperseg')

    if nfft is None:
        nfft = nperseg
    elif nfft < nperseg:
        warnings.warn(
            'nfft < nperseg 可能导致重构不准确。请确保与 STFT 计算时的参数一致。',
            UserWarning,
            stacklevel=2,
        )

    # 计算 hop
    hop = nperseg - noverlap

    # 获取窗函数
    if isinstance(window, str) or isinstance(window, tuple):
        win = get_window(window, nperseg)
    else:
        win = np.asarray(window)

    # 确定 FFT 模式
    fft_mode = 'onesided' if freq_bins == nfft // 2 + 1 else 'twosided'

    # 创建 ShortTimeFFT 对象
    SFT = ShortTimeFFT(win=win, hop=hop, fs=fs, fft_mode=fft_mode, mfft=nfft, **kwargs)

    # 执行 ISTFT
    # 让 ShortTimeFFT 自动确定重构范围
    x = SFT.istft(Zxx, f_axis=freq_axis, t_axis=time_axis)

    # 计算时间数组
    n_samples = len(x) if x.ndim == 1 else x.shape[-1]
    t = np.arange(n_samples) / fs

    return t, x


def spectrogram(
    x,
    fs,
    window='hamming',
    nperseg=None,
    noverlap=None,
    nfft=None,
    axis=-1,
    scaling='density',
    mode='psd',
    **kwargs,
):
    """计算信号的谱图

    计算信号的功率谱图或幅度谱图; 返回 real 类型

    Parameters
    ----------
    - `x` : `array_like`
        - 输入信号。
    - `fs` : `float`
        - 采样频率，单位为 Hz。
    - `window` : `str` or `tuple` or `array_like`, optional
        - 窗函数类型。默认为 'hamming'。
    - `nperseg` : `int`, optional
        - 每段的长度。默认为 `int(0.5 * fs)`。
    - `noverlap` : `int`, optional
        - 段间重叠的样本数。默认为 `nperseg // 2`。
    - `nfft` : `int`, optional
        - FFT 的长度。默认等于 `nperseg`。
    - `axis` : `int`, optional
        - 进行变换的轴。默认为 -1。
    - `scaling` : `str`, optional
        - 缩放模式。'density': 功率谱密度；'spectrum': 功率谱。默认为 'density'。
    - `mode` : `str`, optional
        - 输出模式。'magnitude': 幅度谱；'psd': 功率谱密度；'complex': 复数谱。默认为 'psd'。

    Returns
    ----------
    - `f` : `ndarray`
        - 频率数组，单位为 Hz。
    - `t` : `ndarray`
        - 时间数组，单位为秒。
    - `Sxx` : `ndarray`
        - 谱图。根据 `mode` 参数，可能是幅度谱、功率谱密度或复数谱。
    """
    # 计算 STFT
    f, t, Zxx = stft(
        x=x,
        fs=fs,
        window=window,
        nperseg=nperseg,
        noverlap=noverlap,
        nfft=nfft,
        axis=axis,
        scaling=scaling,
        **kwargs,
    )

    # 根据模式计算输出
    if mode == 'complex':
        Sxx = Zxx
    elif mode == 'magnitude':
        Sxx = np.abs(Zxx)
    elif mode == 'psd':
        Sxx = np.abs(Zxx) ** 2
    else:
        raise ValueError(
            f"不支持的模式: {mode}。支持的模式: 'magnitude', 'psd', 'complex'"
        )

    return f, t, Sxx


def cwt(
    x: np.ndarray,
    fs: float,
    wavelet: str = 'cmor1.5-1.0',
    scales: Optional[np.ndarray] = None,
    n_scales: int = 64,
    fmin: Optional[float] = None,
    fmax: Optional[float] = None,
    axis: int = -1,
    method: Literal['fft', 'conv'] = 'fft',
):
    """连续小波变换 (Continuous Wavelet Transform, CWT)

    对信号进行连续小波变换，特别适用于低频成分重要的脑电信号分析。
    相比STFT，小波变换能提供更好的多分辨率分析能力：低频处有更高的频率分辨率，高频处有更高的时间分辨率。

    Parameters
    ----------
    - `x` : `np.ndarray`
        - 输入信号。支持多维数组，沿指定轴进行变换。
    - `fs` : `float`
        - 采样频率，单位为 Hz。
    - `wavelet` : `str`, optional
        - 小波基函数名称，默认为 'morl'（Morlet小波）。
        - 支持 PyWavelets 库中的所有连续小波函数，包括：
          'cgau1'-'cgau8', 'cmor', 'fbsp', 'gaus1'-'gaus8', 'mexh', 'morl', 'shan'
    - `scales` : `np.ndarray`, optional
        - 小波变换的尺度参数数组。如果为None，根据n_scales, fmin和fmax自动生成
    - `n_scales` : `int`, optional
        - 自动生成的尺度数量，默认为64
    - `fmin` : `float`, optional
        - 最小频率，单位为Hz。默认为None，将被设置为fs/100（即较低频率）
    - `fmax` : `float`, optional
        - 最大频率，单位为Hz。默认为None，将被设置为fs/2（奈奎斯特频率）
    - `axis` : `int`, optional
        - 进行变换的轴。默认为-1（最后一个轴）。
    - `method` : `str`, optional
        - CWT计算方法，默认为 'fft'（FFT方法，速度更快）
        - 'conv': 使用卷积方法（适合短信号）
        - 'fft': 使用FFT方法（适合长信号，速度更快）

    Returns
    ----------
    - `scales` : `np.ndarray`
        - 尺度参数数组
    - `freqs` : `np.ndarray`
        - 对应的频率数组，单位为Hz
    - `t` : `np.ndarray`
        - 时间数组，单位为秒
    - `cwt_data` : `np.ndarray`
        - CWT结果，复数数组。形状为(..., n_scales, n_samples)。

    Examples
    ----------
    ```py
    >>> import numpy as np
    >>> from zyplib.signal.time_frequency import cwt
    >>>
    >>> # 生成测试信号
    >>> fs = 1000
    >>> t = np.arange(0, 3, 1/fs)
    >>> x = np.sin(2*np.pi*10*t) + np.sin(2*np.pi*50*t[:len(t)//2])
    >>>
    >>> # 计算CWT
    >>> scales, freqs, t_cwt, coef = cwt(x, fs, wavelet='cmor', fmin=1, fmax=100)
    >>> print(f"频率范围: {freqs[0]:.1f} - {freqs[-1]:.1f} Hz")
    >>> print(f"时间范围: {t_cwt[0]:.2f} - {t_cwt[-1]:.2f} s")
    ```

    Notes
    -----
    **EEG/MEG 脑电信号分析实践:**

    参数建议: 不设置 `scales`，而是通过设定感兴趣的**频率范围**来让函数自动计算最佳尺度。

    1.  **小波选择 (`wavelet`)**:
        - **Complex Morlet (`cmorB-C`, 如 'cmor1.5-1.0')** 是最常用的选择。它的复数特性使其能同时提供幅度和相位信息，非常适合计算时频功率和锁相值 (phase-locking value)。
        - `B` 是带宽参数，`C` 是中心频率参数。例如，`'cmor1.5-1.0'` 表示带宽为1.5，中心频率为1.0。

    2.  **频率范围 (`fmin`, `fmax`)**:
        - 根据您感兴趣的脑电频带进行设置。
        - **Delta (δ)**: 1-4 Hz
        - **Theta (θ)**: 4-8 Hz
        - **Alpha (α)**: 8-13 Hz
        - **Beta (β)**: 13-30 Hz
        - **Gamma (γ)**: 30-100 Hz
        - 若要进行全面的探索性分析，可设置 `fmin=1`, `fmax=100`。

    3.  **尺度数量 (`n_scales`)**:
        - 该参数控制频率分辨率。`64` 到 `256` 是常用范围。
        - 数量越多，时频图越平滑，但计算成本越高。对于发表级别的图像，通常选择 `128` 或更高。
        - 这个值决定了在`fmin`和`fmax`之间采样的频率点数量。值越大，频率轴越平滑，时频图越精细，但计算也越慢

    4.  **结果解读**:
        - 返回的 `cwt_data` 是复数。
        - **时频功率 (Time-Frequency Power)**: `power = np.abs(cwt_data)**2`。这是最常见的分析，用于观察特定频带能量随时间的变化（即事件相关同步/去同步, ERS/ERD）。
        - **相位 (Phase)**: `phase = np.angle(cwt_data)`。用于分析跨试次或跨通道的相位一致性。
    """
    x = cast(np.ndarray, ensure_npy(x))

    # 将输入数据移动到指定轴
    if axis != -1:
        x = np.moveaxis(x, axis, -1)

    # 计算信号长度
    n_samples = x.shape[-1]

    # 设置默认频率范围
    if fmin is None:
        fmin = fs / 100  # 较低频率，有助于捕捉低频成分
    if fmax is None:
        fmax = fs / 2  # 奈奎斯特频率

    # 确保频率范围有效
    fmin = max(fmin, fs / n_samples)  # 最低频率受信号长度限制
    fmax = min(fmax, fs / 2)  # 最高频率不超过奈奎斯特频率

    # 生成尺度参数
    if scales is None:
        # 将频率转换为尺度（频率与尺度成反比关系）
        # 获取中心频率，用于将尺度转换为频率
        wavelet_obj = cast(Any, pywt).ContinuousWavelet(wavelet)
        central_freq = cast(float, pywt.central_frequency(wavelet_obj))

        # 对数均匀尺度，确保在对数频率轴上均匀分布
        scales = central_freq * fs / np.logspace(np.log10(fmin), np.log10(fmax), n_scales)
        scales = cast(np.ndarray, scales)[::-1]  # 反转顺序，使低频（大尺度）在上方

    # 计算小波变换并获取频率信息
    cwt_data, freqs = pywt.cwt(
        x,
        scales=scales,
        wavelet=wavelet,
        sampling_period=1.0 / fs,
        method=method,
        axis=-1,
    )

    # 确保频率按降序排列（高频在前，与尺度对应）
    if len(freqs) > 1 and freqs[0] < freqs[-1]:
        freqs = freqs[::-1]
        cwt_data = cwt_data[..., ::-1, :]

    # 将结果移回原始轴
    if axis != -1:
        cwt_data = np.moveaxis(cwt_data, -1, axis)

    # 生成时间数组
    t = np.arange(0, n_samples) / fs

    return scales, freqs, t, cwt_data


def dwt(
    x: np.ndarray,
    fs: float,
    wavelet: str = 'db4',
    level: Optional[int] = None,
    mode: str = 'symmetric',
    axis: int = -1,
):
    """离散小波变换 (Discrete Wavelet Transform, DWT)

    对信号进行多分辨率分析，分解为近似系数和多个尺度的细节系数。
    适用于提取信号的时频特征，特别是脑电信号中的不同频段活动。

    Parameters
    ----------
    - `x` : `np.ndarray`
        - 输入信号。支持多维数组，沿指定轴进行变换。
    - `fs` : `float`
        - 采样频率，单位为 Hz。
    - `wavelet` : `str`, optional
        - 小波基函数名称，默认为'db4'（Daubechies 4阶小波）。
        - 常用选择：'db'系列(db1-db38), 'sym'系列, 'coif'系列, 'haar'等
    - `level` : `int`, optional
        - 分解层数。如果为None，则自动选择最大可分解层数。
    - `mode` : `str`, optional
        - 边界处理模式，默认为'symmetric'（对称扩展）。
    - `axis` : `int`, optional
        - 进行变换的轴。默认为-1（最后一个轴）。

    Returns
    ----------
    - `coeffs` : `list`
        - 小波系数列表。[cAn, cDn, cDn-1, ..., cD1]，其中cAn是最后一层的近似系数，cDi是第i层的细节系数。
    - `freqs` : `list`
        - 各层对应的频率带列表，每个元素为(低频界, 高频界)元组，单位为Hz。
    - `t` : `list`
        - 各层对应的时间数组列表，单位为秒。

    Examples
    ----------
    ```py
    >>> import numpy as np
    >>> from zyplib.signal.time_frequency import dwt
    >>>
    >>> # 生成测试信号
    >>> fs = 1000
    >>> t = np.arange(0, 3, 1/fs)
    >>> x = np.sin(2*np.pi*10*t) + 0.5*np.sin(2*np.pi*30*t) + 0.2*np.random.randn(len(t))
    >>>
    >>> # 计算DWT
    >>> coeffs, freqs, t_list = dwt(x, fs, wavelet='db4', level=5)
    >>>
    >>> # 访问第3层的细节系数(对应频带约为62.5-125 Hz)
    >>> d3 = coeffs[-3]
    >>> freq_band = freqs[-3]
    >>> t3 = t_list[-3]
    >>> print(f"第3层细节系数频带: {freq_band[0]:.1f}-{freq_band[1]:.1f} Hz")
    ```

    Notes
    -----
    **EEG/MEG 脑电信号分析实践:**

    DWT 非常适合从EEG信号中**提取频带特征**，用于机器学习模型（如睡眠分期、癫痫检测、BCI等）。

    1.  **小波选择 (`wavelet`)**:
        - **Daubechies (`'db4'`, `'db6'`)** 或 **Symlets (`'sym4'`, `'sym8'`)** 是常用选择
        - 常用推荐: `'db4'` 或 `'sym4'`

    2.  **分解层数 (`level`)**:
        - **这是最关键的参数，必须根据您的采样率 `fs` 来确定。**
        - **计算方法**: 对于第 `k` 层的细节系数 `cDk`，其频带约为 `[fs / 2**(k+1), fs / 2**k]`。
        - **示例 (fs=250 Hz)**:
          - `level=6` 的分解：
            - `cD1`: [62.5, 125] Hz (高Gamma)
            - `cD2`: [31.25, 62.5] Hz (Gamma)
            - `cD3`: [15.625, 31.25] Hz (Beta)
            - `cD4`: [7.81, 15.625] Hz (Alpha)
            - `cD5`: [3.91, 7.81] Hz (Theta)
            - `cD6`: [1.95, 3.91] Hz (Delta)
            - `cA6`: [0, 1.95] Hz (低Delta)
        - 您需要根据您的 `fs` 预先计算好这个对应表。

    3.  **特征提取**:
        - 分解后，`coeffs` 列表包含了每个频带的系数。
        - 常用的特征包括：
          - **能量 (Energy)**: `np.sum(np.square(cD_k))`
          - **功率 (Power)**: `np.mean(np.square(cD_k))`
          - **标准差 (Standard Deviation)**: `np.std(cD_k)`
          - **熵 (Entropy)**: `scipy.stats.entropy(np.square(cD_k))`
        - 将这些特征组合起来，可以构成用于分类器的特征向量。
    """
    x = cast(np.ndarray, ensure_npy(x))

    # 将输入数据移动到指定轴
    if axis != -1:
        x = np.moveaxis(x, axis, -1)

    # 计算信号长度
    n_samples = x.shape[-1]

    # 自动确定最大分解层数
    max_level = pywt.dwt_max_level(
        data_len=n_samples, filter_len=cast(Any, pywt).Wavelet(wavelet).dec_len
    )
    if level is None:
        level = max_level
    else:
        level = min(level, max_level)  # 不超过最大可分解层数
    level = cast(int, level)

    # 计算小波分解
    coeffs = pywt.wavedec(x, wavelet=wavelet, level=level, mode=mode, axis=-1)

    # 计算各层的频率带
    freqs = []
    current_fs = fs

    # 对于近似系数（最低频带）
    low_freq = 0
    high_freq = current_fs / 2 ** (level + 1)
    freqs.append((low_freq, high_freq))

    # 对于各层细节系数（从最深层到浅层）
    for i in range(level):
        # 计算当前层的频率带
        low_freq = current_fs / 2 ** (level - i + 1)
        high_freq = current_fs / 2 ** (level - i)
        freqs.append((low_freq, high_freq))

    # 计算各层的时间数组
    t_list = []
    for i, coef in enumerate(coeffs):
        # 获取当前层系数的长度
        coef_len = coef.shape[-1]
        # 计算时间步长（较深层的时间分辨率更低）
        # 计算每层的下采样因子
        downsample_factor = 2 ** (level - i + 1) if i > 0 else 2**level
        # 计算时间步长
        dt = downsample_factor / fs
        t_layer = np.arange(coef_len) * dt
        t_list.append(t_layer)

    return coeffs, freqs, t_list



"""
本模块主要提供了一些数据集导入的 pipeline 相关的函数，是对脑电信号数据导入、预处理流程的高级封装

所有函数均满足:

- 输入: 一些参数，以及一些回调函数
- 输出: 返回数据集 (X, y)

"""

import typing

from functools import reduce
from typing import Callable, Literal, Protocol, Sequence, cast, runtime_checkable

import numpy as np
from sklearn.model_selection import KFold

from zyplib.dataset.split import random_split
from zyplib.preprocessing.labels import to_onehot
from zyplib.utils.print import print_info
from zyplib.utils.progress_bar import rich_progress

__all__ = [
    'fileseg_pipeline',
    'fileloader_pipeline',
    'basic_pipeline',
    'simple_pipeline',
    'train_test_pipeline',
    'kfold_pipeline',
]


@runtime_checkable
class FuncPipeline(Protocol):
    """函数类型，输入文件路径列表，输出数据集 (X, y)"""

    def __call__(self, files: typing.Sequence[str]) -> tuple[np.ndarray, np.ndarray]: ...


@runtime_checkable
class ReadFunction(Protocol):
    def __call__(self, file_path: str) -> np.ndarray: ...


@runtime_checkable
class SegmentFunction(Protocol):
    def __call__(self, signal: np.ndarray) -> np.ndarray: ...


@runtime_checkable
class LabelFunction(Protocol):
    def __call__(self, file_path: str) -> int | list[int]: ...


@runtime_checkable
class SignalProcessFunction(Protocol):
    def __call__(self, signal: np.ndarray) -> np.ndarray: ...


@runtime_checkable
class EncodeLabelFunction(Protocol):
    def __call__(self, label: typing.Sequence[int] | np.ndarray) -> np.ndarray: ...


@runtime_checkable
class TransformFunction(Protocol):
    def __call__(self, file_path: str) -> tuple[np.ndarray, np.ndarray]: ...


def _default_encode_label(label: typing.Sequence[int] | np.ndarray) -> np.ndarray:
    return np.asarray(to_onehot(label, squeeze_2_class=True))


def _identity_signal(signal: np.ndarray) -> np.ndarray:
    return signal


def _normalize_file_groups(
    all_files: typing.Sequence[str] | Sequence[typing.Sequence[str]],
) -> list[list[str]]:
    if not all_files:
        raise ValueError("The 'all_files' list cannot be empty.")

    first_group = all_files[0]
    if not first_group:
        raise ValueError("The 'all_files' list cannot be empty.")

    if isinstance(first_group, str):
        flat_files = cast(typing.Sequence[str], all_files)
        return [sorted(flat_files)]

    grouped_files = cast(Sequence[typing.Sequence[str]], all_files)
    return [sorted(group) for group in grouped_files]


def fileseg_pipeline(
    files: typing.Sequence[str],
    fn_read: ReadFunction,
    label: int | LabelFunction,
    fn_segment: SegmentFunction | None = None,
    fn_before_segment: SignalProcessFunction | None = None,
    fn_after_segment: SignalProcessFunction | None = None,
    fn_encode_label: EncodeLabelFunction | None = _default_encode_label,
    show_progress: bool = True,
    random_shuffle: bool | int = False,
) -> tuple[np.ndarray, np.ndarray]:
    """从文件中读取数据，并自动切片、制作用于训练的数据集

    本 Pipeline 适用于:
    - 从文件中读取完整的脑电
    - 需要对完整脑电进行切分
    - 无特殊标签，只需根据文件进行标记即可
    - 输出: 数据集 (X, y)
        - X: 数据，shape = (N, C, T)
        - y: 标签，shape = (N,)

    Pipeline 流程:
    ----------

    1. 从文件中读取完整的数据 X
    2. 对数据 X 进行预处理
    3. 对数据 X 进行切片
    4. 对数据 X 进行后处理
    5. 制作标签 y
    6. 返回数据集 (X, y)

    Parameters
    ----------
    - `files` : `list[str]`
        - 文件路径列表
    - `fn_read` : `(file_path: str) -> np.ndarray`
        - 文件读取函数
    - `fn_segment` : `(signal: np.ndarray) -> np.ndarray`
        - 数据切片函数
        - 输入一个 `[C, T]` 的信号
        - 输出一个 `[N, C, T]` 的信号
    - `label` : `int | (file_path: str) -> int`
        - 标签，如果为 int，则所有数据标签相同；如果为函数，则根据文件名生成标签
    - `fn_before_segment` : `(signal: [C, T]) -> [C, T]`, optional
        - 数据切片前的处理函数
    - `fn_after_segment` : `(signal: [N, C, T]) -> [N, C, T]`, optional
        - 数据切片后的处理函数
    - `fn_encode_label` : `(label: list | np.ndarray) -> np.ndarray`, optional
        - 标签编码函数, 默认使用 `to_onehot`
    - `show_progress` : `bool`, optional
        - 是否显示处理进度, by default `True`
    - `random_shuffle` : `bool | int`, optional
        - 是否在返回前随机打乱数据
        - `False`: 不打乱数据
        - `True`: 使用默认随机种子打乱数据
        - `int`: 使用指定的整数作为随机种子打乱数据
        - 注意: 此参数不会影响全局随机状态

    Returns
    ----------
    - `X` : `np.ndarray`, shape = (N, C, T)
        - 数据
    - `y` : `np.ndarray`, shape = (N, n_classes)
        - 标签
    """
    if not files:
        raise ValueError('The files list is empty. Please provide at least one file.')

    segment_fn: SegmentFunction = (
        fn_segment if fn_segment is not None else _identity_signal
    )
    before_segment_fn: SignalProcessFunction = (
        fn_before_segment if fn_before_segment is not None else _identity_signal
    )
    after_segment_fn: SignalProcessFunction = (
        fn_after_segment if fn_after_segment is not None else _identity_signal
    )
    fn_label = label if callable(label) else (lambda _: label)

    # 处理单个文件
    def load_single_file(fpath: str):
        signal = fn_read(fpath)  # 读取数据
        signal = before_segment_fn(signal)  # 数据预处理
        segments = segment_fn(signal)  # 数据切片
        segments = after_segment_fn(segments)
        N = len(segments)
        label_ = fn_label(fpath)  # 制作标签
        labels = []
        if isinstance(label_, list):
            labels = label_
        else:
            labels = [label_] * N
        return segments, labels

    # 批量处理所有文件
    if show_progress:
        import os

        iterable = rich_progress(
            files, description='处理中', fn_get_status=lambda f: os.path.basename(f)
        )
    else:
        iterable = files
    results = map(load_single_file, iterable)

    # 合并数据
    def merge_data(
        Xy: tuple[list[np.ndarray], list[int]],
        pair: tuple[np.ndarray, typing.Sequence[int]],
    ):
        segments, labels = pair
        if len(segments) > 0:
            Xy[0].append(segments)
            Xy[1].extend(labels)
        return Xy

    Xy = reduce(merge_data, results, ([], []))

    if not Xy[0]:
        X = np.array([])
        y = np.array([], dtype=int)
    else:
        X = np.concatenate(Xy[0], axis=0)
        y = np.array(Xy[1], dtype=int)

    encode_label_fn = fn_encode_label or _default_encode_label
    y = encode_label_fn(y)

    # 随机打乱数据
    if random_shuffle is not False:
        # 创建独立的随机数生成器实例
        if isinstance(random_shuffle, int):
            # 使用指定的种子
            rand_gen = np.random.RandomState(seed=random_shuffle)
        elif random_shuffle is True:
            # 使用默认种子
            rand_gen = np.random.RandomState()
        else:
            raise ValueError('random_shuffle must be a boolean or an integer')

        # 生成随机索引并打乱数据
        if len(X) > 0:  # 确保数据不为空
            indices = rand_gen.permutation(len(X))
            X = X[indices]
            y = y[indices]

    return X, y


basic_pipeline = fileseg_pipeline


def fileloader_pipeline(
    files: typing.Sequence[str],
    fn_fileloader: TransformFunction,
    fn_encode_label: EncodeLabelFunction | None = _default_encode_label,
    show_progress: bool = True,
    random_shuffle: bool | int = False,
) -> tuple[np.ndarray, np.ndarray]:
    """从文件中读取，制作用于训练的数据集

    本函数相对 fileseg_pipeline 更加灵活，核心是 fn_fileloader 函数，可以用于处理数据并制作数据集

    ```py
    fn_fileloader(file_path) -> (Data, Label)
    ```

    Pipeline 流程:
    ----------

    1. 从每个文件中读取 (X, y) 数据
    2. 合并所有数据
    3. 对标签进行编码
    4. 返回数据集 (X, y)

    Parameters
    ----------
    - `files` : `list[str]`
        - 文件路径列表
    - `fn_fileloader` : `(file_path: str) -> (np.ndarray, np.ndarray)`
        - 文件变换函数，输入文件路径，输出 (X, y) 数据
        - X: shape = (N, C, T)，数据
        - y: shape = (N,)，标签
    - `fn_encode_label` : `(label: list | np.ndarray) -> np.ndarray`, optional
        - 标签编码函数, 默认使用 `to_onehot`
    - `show_progress` : `bool`, optional
        - 是否显示处理进度, by default `True`
    - `random_shuffle` : `bool | int`, optional
        - 是否在返回前随机打乱数据
        - `False`: 不打乱数据
        - `True`: 使用默认随机种子打乱数据
        - `int`: 使用指定的整数作为随机种子打乱数据
        - 注意: 此参数不会影响全局随机状态

    Returns
    ----------
    - `X` : `np.ndarray`, shape = (N, C, T)
        - 数据
    - `y` : `np.ndarray`, shape = (N, n_classes)
        - 标签
    """
    if not files:
        raise ValueError('The files list is empty. Please provide at least one file.')

    # 处理单个文件
    def load_single_file(fpath: str):
        X, y = fn_fileloader(fpath)  # 读取数据
        return X, y

    # 批量处理所有文件
    if show_progress:
        import os

        iterable = rich_progress(
            files, description='处理中', fn_get_status=lambda f: os.path.basename(f)
        )
    else:
        iterable = files
    results = map(load_single_file, iterable)

    # 合并数据
    def merge_data(
        Xy: tuple[list[np.ndarray], list[np.ndarray]], pair: tuple[np.ndarray, np.ndarray]
    ):
        X_batch, y_batch = pair
        if len(X_batch) > 0:
            Xy[0].append(X_batch)
            Xy[1].append(y_batch)
        return Xy

    Xy = reduce(merge_data, results, ([], []))

    if not Xy[0]:
        X = np.array([])
        y = np.array([], dtype=int)
    else:
        X = np.concatenate(Xy[0], axis=0)
        y = np.concatenate(Xy[1], axis=0)

    encode_label_fn = fn_encode_label or _default_encode_label
    y = encode_label_fn(y)

    # 随机打乱数据
    if random_shuffle is not False:
        # 创建独立的随机数生成器实例
        if isinstance(random_shuffle, int):
            # 使用指定的种子
            rand_gen = np.random.RandomState(seed=random_shuffle)
        elif random_shuffle is True:
            # 使用默认种子
            rand_gen = np.random.RandomState()
        else:
            raise ValueError('random_shuffle must be a boolean or an integer')

        # 生成随机索引并打乱数据
        if len(X) > 0:  # 确保数据不为空
            indices = rand_gen.permutation(len(X))
            X = X[indices]
            y = y[indices]

    return X, y


simple_pipeline = fileloader_pipeline  # Alias


def train_test_pipeline(
    split_size: float | int | tuple,
    all_files: typing.Sequence[str] | Sequence[typing.Sequence[str]],
    random_seed: int | None = None,
    return_split_files: bool = False,
    *,
    fn_pipeline: FuncPipeline
    | Literal['fileseg', 'fileloader', 'basic', 'simple'] = 'fileseg',
    pipeline_kwargs: typing.Mapping | None = None,
):
    """在导入文件的时候，自动将不同文件划分为训练集、验证集和测试集

    `split_size` 控制划分方式:
    - `number`: 视为训练集大小，剩余为测试集 (2-way split)。
    - `(train, test)`: 指定训练集和测试集大小 (2-way split)。
    - `(train, val, test)`: 指定训练、验证和测试集大小 (3-way split)。
    大小可以是 `int` (绝对数量) 或 `float` (比例)。

    `fn_pipeline` 回调函数: `(list[str]) -> (np.ndarray, np.ndarray)`
    - 输入: 文件路径列表
    - 输出: 数据集 (X, y)
    当 `fn_pipeline` 为字符串时，使用对应的预设 pipeline。

    #### 使用示例

    ```py
    # 2-way split with ratio
    X_train, X_test, y_train, y_test = train_test_pipeline(
        0.8, all_files, fn_pipeline=my_load_xy
    )

    # 3-way split with ratio
    X_train, X_val, X_test, y_train, y_val, y_test = train_test_pipeline(
        (0.7, 0.1, 0.2), all_files, fn_pipeline=my_load_xy
    )
    ```

    Parameters
    ----------
    - `split_size` : `float | int | tuple`
        - 数据集划分大小或比例。
        - `number`: 训练集大小，剩余为测试集。
        - `(train, test)`: 训练集和测试集大小。
        - `(train, val, test)`: 训练集、验证集和测试集大小。
    - `all_files` : `list[str] | Sequence[list[str]]`
        - 所有文件路径列表。
        - `list[str]`: 直接对整个文件列表进行划分。
        - `Sequence[list[str]]`: 按组进行分层划分。每个子列表被视为一个独立的组（例如，一个类别），
          切分比例将独立应用于每个组，以确保在训练/测试集中保持类别平衡。
    - `random_seed` : `int | None`, optional
        - 随机种子; 当 `random_seed` 为 `None` 时，不设置随机种子。
    - `return_split_files` : `bool`, optional
        - 是否返回划分后的文件路径列表。
    - `fn_pipeline` : `FuncPipeline | Literal['fileseg', 'fileloader', 'basic', 'simple']`, optional
        - 数据处理 pipeline; 默认为 `'fileseg'`，使用 `fileseg_pipeline`。
        - `'fileseg'` / `'basic'`: 使用 `fileseg_pipeline`（信号切片流程）
        - `'fileloader'` / `'simple'`: 使用 `fileloader_pipeline`（灵活加载流程）
    - `**pipeline_kwargs` :
        - pipeline 函数的参数
        - 当使用 `'fileseg'` / `'basic'` 时，包括: `['fn_read', 'fn_segment', 'label', 'fn_before_segment', 'fn_after_segment', 'show_progress', 'fn_encode_label', 'random_shuffle']`
        - 当使用 `'fileloader'` / `'simple'` 时，包括: `['fn_fileloader', 'fn_encode_label', 'show_progress', 'random_shuffle']`

    Returns
    ----------
    - `*data`: `np.ndarray`
        - 划分后的数据集。根据 `split_size` 的不同，返回 `(X_train, X_test, y_train, y_test)` 或 `(X_train, X_val, X_test, y_train, y_val, y_test)`。
    - `*files`: `list[str]`, optional
        - 划分后的文件列表。当 `return_split_files` 为 `True` 时返回。
    """
    all_files = _normalize_file_groups(all_files)

    train_files, val_files, test_files = [], [], []
    has_val = isinstance(split_size, tuple) and len(split_size) == 3

    for group in all_files:
        N = len(group)
        sizes = []
        if isinstance(split_size, (int, float)):
            train_s = int(N * split_size) if isinstance(split_size, float) else split_size
            test_s = N - train_s
            sizes = (train_s, test_s)
        elif isinstance(split_size, tuple):
            raw_sizes = [int(N * s) if isinstance(s, float) else s for s in split_size]
            if isinstance(split_size[-1], float):
                raw_sizes[-1] = N - sum(raw_sizes[:-1])
            sizes = tuple(raw_sizes)
        else:
            raise TypeError(f'Unsupported type for split_size: {type(split_size)}')

        if sum(sizes) > N:
            raise ValueError(
                f'Sum of split sizes ({sum(sizes)}) cannot exceed the number of files ({N}).'
            )

        splits = random_split(group, sizes, seed=random_seed)

        train_files.extend(splits[0])
        if has_val:
            val_files.extend(splits[1])
            test_files.extend(splits[2])
        else:
            test_files.extend(splits[1])

    pipeline_kwargs = pipeline_kwargs or {}
    if fn_pipeline in ('fileseg', 'basic'):
        fn_pipeline = lambda files: fileseg_pipeline(files, **pipeline_kwargs)
    elif fn_pipeline in ('fileloader', 'simple'):
        fn_pipeline = lambda files: fileloader_pipeline(files, **pipeline_kwargs)

    def process_files(file_list: list[str], name: str) -> tuple[np.ndarray, np.ndarray]:
        if not file_list:
            return np.array([]), np.array([])
        print_info(f'Processing {name} files...')
        return fn_pipeline(file_list)

    X_train, y_train = process_files(train_files, 'train')
    X_test, y_test = process_files(test_files, 'test')

    if has_val:
        X_val, y_val = process_files(val_files, 'val')
        data_to_return = (X_train, X_val, X_test, y_train, y_val, y_test)
        files_to_return = (train_files, val_files, test_files)
    else:
        data_to_return = (X_train, X_test, y_train, y_test)
        files_to_return = (train_files, test_files)

    if return_split_files:
        return (*data_to_return, *files_to_return)
    else:
        return data_to_return


def kfold_pipeline(
    k: int,
    all_files: typing.Sequence[str] | Sequence[typing.Sequence[str]],
    random_seed: int | None = None,
    yield_split_files: bool = False,
    *,
    fn_pipeline: FuncPipeline
    | Literal['fileseg', 'fileloader', 'basic', 'simple'] = 'fileseg',
    pipeline_kwargs: typing.Mapping | None = None,
):
    """在导入文件的时候，自动将不同文件划分为 k 折, 通过 `yield` 返回每次的 XY 数据

    - 需要一个 `fn_pipeline: (list[str]) -> (np.ndarray, np.ndarray)` 的回调函数
        - 输入: 文件路径列表
        - 输出: 数据集 (X, y)
    - 当 `fn_pipeline` 为字符串时，使用对应的预设 pipeline，并使用 `**pipeline_kwargs` 作为参数

    Parameters
    ----------
    - `k` : `int`
        - 折数
    - `all_files` : `list[str] | Sequence[list[str]]`
        - 所有文件路径列表。
        - `list[str]`: 直接对整个文件列表进行 K 折交叉验证。
        - `Sequence[list[str]]`: 按组进行分层 K 折交叉验证。每个子列表被视为一个独立的组，
          K 折切分将独立应用于每个组，然后将每折的结果合并，以确保类别平衡。
    - `random_seed` : `int | None`, optional
        - 随机种子; 当 `random_seed` 为 `None` 时，不设置随机种子
    - `yield_split_files` : `bool`, optional
        - 是否返回训练集和验证集的文件路径列表; 当 `yield_split_files` 为 `True` 时，返回训练集和验证集的文件路径列表
    - `fn_pipeline` : `FuncPipeline | Literal['fileseg', 'fileloader', 'basic', 'simple']`, optional
        - 数据处理 pipeline; 默认为 `'fileseg'`，使用 `fileseg_pipeline`
        - `'fileseg'` / `'basic'`: 使用 `fileseg_pipeline`（信号切片流程）
        - `'fileloader'` / `'simple'`: 使用 `fileloader_pipeline`（灵活加载流程）
    - `**pipeline_kwargs` :
        - pipeline 函数的参数
        - 当使用 `'fileseg'` / `'basic'` 时，包括: `['fn_read', 'fn_segment', 'label', 'fn_before_segment', 'fn_after_segment', 'show_progress', 'fn_encode_label', 'random_shuffle']`
        - 当使用 `'fileloader'` / `'simple'` 时，包括: `['fn_fileloader', 'fn_encode_label', 'show_progress', 'random_shuffle']`

    Yields
    ----------
    - `X_train` : `np.ndarray`, shape = (N_train, C, T)
        - 训练数据
    - `X_val` : `np.ndarray`, shape = (N_val, C, T)
        - 验证数据
    - `y_train` : `np.ndarray`, shape = (N_train, n_classes)
        - 训练标签
    - `y_val` : `np.ndarray`, shape = (N_val, n_classes)
        - 验证标签
    - `train_files` : `list[str]`, optional
        - 训练集文件路径列表; 当 `yield_split_files` 为 `True` 时返回
    - `val_files` : `list[str]`, optional
        - 验证集文件路径列表; 当 `yield_split_files` 为 `True` 时返回
    """
    all_files = _normalize_file_groups(all_files)

    # 验证每个组的文件数量
    for group in all_files:
        if len(group) < k:
            raise ValueError(
                f'Number of files in a group ({len(group)}) is less than k ({k}). '
                'Cannot perform K-Fold splitting.'
            )

    # 为每个组预先生成所有的分割索引
    all_splits = []
    for group in all_files:
        kf = KFold(n_splits=k, shuffle=True, random_state=random_seed)
        splits = list(kf.split(group))  # 获取所有 k 折的分割索引
        all_splits.append(splits)

    pipeline_kwargs = pipeline_kwargs or {}
    if fn_pipeline in ('fileseg', 'basic'):
        fn_pipeline = lambda files: fileseg_pipeline(files, **pipeline_kwargs)
    elif fn_pipeline in ('fileloader', 'simple'):
        fn_pipeline = lambda files: fileloader_pipeline(files, **pipeline_kwargs)

    for fold_idx in range(k):
        print_info(f'Processing fold {fold_idx + 1}')
        train_files, val_files = [], []

        # 从每个组的预生成分割中获取当前折的索引
        for i, group in enumerate(all_files):
            train_idx, val_idx = all_splits[i][fold_idx]
            train_files.extend([group[j] for j in train_idx])
            val_files.extend([group[j] for j in val_idx])

        print_info('Processing train files...')
        X_train, y_train = fn_pipeline(train_files)
        print_info('Processing val files...')
        X_val, y_val = fn_pipeline(val_files)

        if yield_split_files:
            yield X_train, X_val, y_train, y_val, train_files, val_files
        else:
            yield X_train, X_val, y_train, y_val



import typing
import copy
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Literal,
    Optional,
    Sequence,
    TypeAlias,
    TypeVar,
    Union,
)

from lightning.fabric import Fabric
from lightning.fabric.wrappers import _FabricModule, _FabricOptimizer
from torch import nn
from torch.optim import Optimizer

from zyplib.nn.utils import load_checkpoint
from zyplib.train.common import callbacks
from zyplib.train.metric.torch_metric import TorchMetricRecorder
from zyplib.utils.print import print_colored, warning

if TYPE_CHECKING:
    from torch import Tensor
    from torch.optim.lr_scheduler import _LRScheduler
    from torch.utils.data import DataLoader

# 类型定义
TensorOrTensors: TypeAlias = Union['Tensor', Sequence['Tensor']]
Batch: TypeAlias = tuple['Tensor', ...]
Metrics: TypeAlias = Dict[str, float]
SingleOptimizer: TypeAlias = Optimizer | _FabricOptimizer
OptimizerGroup: TypeAlias = (
    SingleOptimizer
    | tuple[SingleOptimizer, ...]
    | list[SingleOptimizer]
    | dict[str, SingleOptimizer]
)
T = TypeVar('T')


@dataclass
class FabricTrainingContext:
    """训练上下文对象，用于存储训练相关的状态和配置

    提供了对 Lightning Fabric 回调函数的封装，使得在不使用 train_loop 时也能方便地复用回调函数工具。

    Attributes:
    ----------
    - `model` : `nn.Module | _FabricModule`
        - 模型
    - `optimizer` : `Optimizer | tuple[Optimizer, ...] | list[Optimizer] | dict[str, Optimizer]`
        - 优化器，支持单个优化器或多个优化器的组合形式
    - `fabric` : `Fabric | None`, optional
        - 可选项, 如果传入了, 则会调用 fabric 的 `backward` 和 `callback`
    - `loss_fn` : `nn.Module | None`, optional
        - 损失函数，如果传入了则会作为默认损失函数使用，默认为 None
        - 可以通过函数参数动态设置和更新
    - `train_metric_recorder` : `TorchMetricRecorder | None`, optional
        - 用于记录训练时指标的 recorder 实例，默认为 None
    - `val_metric_recorder` : `TorchMetricRecorder | None`, optional
        - 用于记录验证时指标的 recorder 实例，默认为 None

    Methods:
    --------
    回调函数封装方法：
    - `call_on_train_start()` - 训练开始时的回调
    - `call_on_train_end()` - 训练结束时的回调
    - `call_on_epoch_start()` - 每个epoch开始时的回调
    - `call_on_epoch_end()` - 每个epoch结束时的回调
    - `call_on_train_batch_end()` - 每个训练批次结束时的回调
    - `call_on_validation_batch_end()` - 每个验证批次结束时的回调
    - `call_on_test_start()` - 测试开始时的回调
    - `call_on_test_end()` - 测试结束时的回调
    - 以及其他相关的回调函数封装...
    """

    model: nn.Module | _FabricModule
    optimizer: OptimizerGroup
    fabric: Optional[Fabric] = None
    fn_loss: Optional[nn.Module] = None
    train_metric_recorder: Optional[TorchMetricRecorder] = None
    val_metric_recorder: Optional[TorchMetricRecorder] = None

    @property
    def pure_model(self) -> nn.Module:
        """返回 unwrap 后的原始模型"""

        if self.fabric and isinstance(self.model, _FabricModule):
            return self.model.module
        return self.model

    @property
    def pure_optimizer(
        self,
    ) -> Optimizer | typing.Iterator[Optimizer] | dict[str, Optimizer]:
        """返回 unwrap 后的原始优化器"""

        def unwrap(optimizer: SingleOptimizer) -> Optimizer:
            if isinstance(optimizer, _FabricOptimizer):
                return optimizer._optimizer
            return optimizer

        if isinstance(self.optimizer, (tuple, list)):
            return (unwrap(o) for o in self.optimizer)
        elif isinstance(self.optimizer, dict):
            return {k: unwrap(v) for k, v in self.optimizer.items()}
        else:
            return unwrap(self.optimizer)

    def _require_fabric(self) -> Fabric:
        return typing.cast(Fabric, self.fabric)

    def __post_init__(self):
        if self.fabric is None:
            return

        # 检查模型是否已 setup
        if not isinstance(self.model, _FabricModule):
            self.model = self.fabric.setup(self.model)

        # 检查优化器是否已 setup
        if isinstance(self.optimizer, Optimizer) and (
            not isinstance(self.optimizer, _FabricOptimizer)
        ):
            # 如果是单个优化器
            self.optimizer = typing.cast(
                SingleOptimizer, self.fabric.setup_optimizers(self.optimizer)
            )
        elif isinstance(self.optimizer, (list, tuple)):
            # 如果是列表/元组，检查其中每个优化器
            are_optimizers_setup = all(
                isinstance(opt, _FabricOptimizer) for opt in self.optimizer
            )
            if not are_optimizers_setup:
                raw_optimizers = [
                    opt._optimizer if isinstance(opt, _FabricOptimizer) else opt
                    for opt in self.optimizer
                ]
                self.optimizer = typing.cast(
                    OptimizerGroup, self.fabric.setup_optimizers(*raw_optimizers)
                )
        elif isinstance(self.optimizer, dict):
            for key, optimizer in self.optimizer.items():
                if isinstance(optimizer, Optimizer) and not isinstance(
                    optimizer, _FabricOptimizer
                ):
                    self.optimizer[key] = typing.cast(
                        SingleOptimizer, self.fabric.setup_optimizers(optimizer)
                    )

    def launch(self):
        fabric = self._require_fabric()
        fabric.launch()
        if isinstance(self.train_metric_recorder, TorchMetricRecorder):
            self.train_metric_recorder.to_device(fabric.device)
        if isinstance(self.val_metric_recorder, TorchMetricRecorder):
            self.val_metric_recorder.to_device(fabric.device)

    def find_best_ckpt(self):
        fabric = self._require_fabric()
        save_ckpt_cb: callbacks.SaveCheckpoints | None = callbacks.find_callback(
            fabric, callbacks.SaveCheckpoints
        )
        if not save_ckpt_cb:
            warning('当前训练没有注册 SaveCheckpoints 回调函数, 无法加载最佳模型')
            return None
        return save_ckpt_cb.best_checkpoint()

    def load_best_ckpt(self, model: Literal['replace', 'return'] = 'replace'):
        """导入最佳模型

        Parameters
        ----------
        - `model` : `Literal['replace', 'return']`
            - 'replace': 直接替换当前模型
            - 'return': 返回另一个加载后的模型
        """
        path = self.find_best_ckpt()
        print_colored('magenta', f'Load from best checkpoint path: {path}')
        if not path:
            warning('当前训练没有注册 SaveCheckpoints 回调函数, 无法加载最佳模型')
            return None
        fabric = self._require_fabric()
        if model == 'replace':
            load_checkpoint(self.model, str(path), device=fabric.device)
            return self.model
        elif model == 'return':
            new_model = copy.deepcopy(self.pure_model)
            load_checkpoint(new_model, str(path), device=fabric.device)
            return new_model

    # ===================== 回调函数封装 ===================== #

    def call_on_train_start(
        self,
        *,
        lr_scheduler: Optional['_LRScheduler'] = None,
        max_epochs: Optional[int] = None,
        dataloader: Optional['DataLoader'] = None,
        hparams: Optional[typing.Mapping[str, Any]] = None,
        **kwargs,
    ):
        """封装 on_train_start 回调"""
        if self.fabric:
            self.fabric.call(
                'on_train_start',
                model=self.pure_model,
                optimizer=self.pure_optimizer,
                lr_scheduler=lr_scheduler,
                fabric=self.fabric,
                max_epochs=max_epochs,
                dataloader=dataloader,
                hparams=hparams,
                **kwargs,
            )

    def call_on_train_end(self, **kwargs):
        """封装 on_train_end 回调"""
        if self.fabric:
            self.fabric.call('on_train_end', **kwargs)

    def call_on_test_start(self, *, dataloader: Optional['DataLoader'] = None, **kwargs):
        """封装 on_test_start 回调"""
        if self.fabric:
            self.fabric.call('on_test_start', dataloader=dataloader, **kwargs)

    def call_on_test_end(
        self, *, metrics: Optional[typing.Mapping[str, float]] = None, **kwargs
    ):
        """封装 on_test_end 回调"""
        if self.fabric:
            self.fabric.call('on_test_end', metrics=metrics, **kwargs)

    # --------------- Epoch wise --------------- #

    def call_on_epoch_start(self, epoch: int, max_epochs: int, **kwargs):
        """封装 on_epoch_start 回调"""
        if self.fabric:
            self.fabric.call(
                'on_epoch_start', epoch=epoch, max_epochs=max_epochs, **kwargs
            )

    def call_on_epoch_end(
        self,
        epoch: int,
        *,
        train_metrics: Optional[typing.Mapping[str, float]] = None,
        val_metrics: Optional[typing.Mapping[str, float]] = None,
        **kwargs,
    ):
        """封装 on_epoch_end 回调"""
        if self.fabric:
            self.fabric.call(
                'on_epoch_end',
                epoch=epoch,
                train_metrics=train_metrics,
                val_metrics=val_metrics,
                **kwargs,
            )

    def call_on_train_epoch_start(self, epoch: int, max_epochs: int, **kwargs):
        """封装 on_train_epoch_start 回调"""
        if self.fabric:
            self.fabric.call(
                'on_train_epoch_start', epoch=epoch, max_epochs=max_epochs, **kwargs
            )

    def call_on_train_epoch_end(
        self,
        epoch: int,
        *,
        metrics: Optional[typing.Mapping[str, float]] = None,
        **kwargs,
    ):
        """封装 on_train_epoch_end 回调"""
        if self.fabric:
            self.fabric.call('on_train_epoch_end', epoch=epoch, metrics=metrics, **kwargs)

    def call_on_validation_epoch_start(
        self, epoch: int, *, epoch_interval: Optional[int] = None, **kwargs
    ):
        """封装 on_validation_epoch_start 回调"""
        if self.fabric:
            self.fabric.call(
                'on_validation_epoch_start',
                epoch=epoch,
                epoch_interval=epoch_interval,
                **kwargs,
            )

    def call_on_validation_epoch_end(
        self, epoch: int, *, metrics: typing.Mapping[str, float], **kwargs
    ):
        """封装 on_validation_epoch_end 回调"""
        if self.fabric:
            self.fabric.call(
                'on_validation_epoch_end', epoch=epoch, metrics=metrics, **kwargs
            )

    # --------------- Batch wise --------------- #

    def call_on_train_batch_start(
        self, batch: 'TensorOrTensors', batch_idx: int, **kwargs
    ):
        """封装 on_train_batch_start 回调"""
        if self.fabric:
            self.fabric.call(
                'on_train_batch_start', batch=batch, batch_idx=batch_idx, **kwargs
            )

    def call_on_train_batch_end(
        self,
        batch: 'TensorOrTensors',
        batch_idx: int,
        *,
        metrics: Optional[typing.Mapping[str, float]] = None,
        **kwargs,
    ):
        """封装 on_train_batch_end 回调"""
        if self.fabric:
            self.fabric.call(
                'on_train_batch_end',
                batch=batch,
                batch_idx=batch_idx,
                metrics=metrics,
                **kwargs,
            )

    def call_on_validation_batch_start(
        self, batch: 'TensorOrTensors', batch_idx: int, **kwargs
    ):
        """封装 on_validation_batch_start 回调"""
        if self.fabric:
            self.fabric.call(
                'on_validation_batch_start', batch=batch, batch_idx=batch_idx, **kwargs
            )

    def call_on_validation_batch_end(
        self,
        batch: 'TensorOrTensors',
        batch_idx: int,
        *,
        metrics: Optional[typing.Mapping[str, float]] = None,
        **kwargs,
    ):
        """封装 on_validation_batch_end 回调"""
        if self.fabric:
            self.fabric.call(
                'on_validation_batch_end',
                batch=batch,
                batch_idx=batch_idx,
                metrics=metrics,
                **kwargs,
            )

    def call_on_test_batch_start(
        self, batch: 'TensorOrTensors', batch_idx: int, **kwargs
    ):
        """封装 on_test_batch_start 回调"""
        if self.fabric:
            self.fabric.call(
                'on_test_batch_start', batch=batch, batch_idx=batch_idx, **kwargs
            )

    def call_on_test_batch_end(
        self,
        batch: 'TensorOrTensors',
        batch_idx: int,
        *,
        metrics: Optional[typing.Mapping[str, float]] = None,
        **kwargs,
    ):
        """封装 on_test_batch_end 回调"""
        if self.fabric:
            self.fabric.call(
                'on_test_batch_end',
                batch=batch,
                batch_idx=batch_idx,
                metrics=metrics,
                **kwargs,
            )

    # ===================== Transform Hooks ===================== #

    def transform_on_train_batch_start(
        self, batch: 'TensorOrTensors', batch_idx: int, epoch_idx: Optional[int] = None
    ) -> 'TensorOrTensors':
        """训练批次开始前的数据转换 hook（默认实现：恒等变换）

        可在子类中重写以实现数据增强（如 Mixup, CutMix）

        Parameters
        ----------
        - `batch` : `TensorOrTensors`
            - 原始批次数据
        - `batch_idx` : `int`
            - 批次索引
        - `epoch_idx` : `Optional[int]`, optional
            - 当前 epoch 索引

        Returns
        -------
        - `TensorOrTensors`
            - 转换后的批次数据
        """
        return batch

    def transform_on_validation_batch_start(
        self, batch: 'TensorOrTensors', batch_idx: int, epoch_idx: Optional[int] = None
    ) -> 'TensorOrTensors':
        """验证批次开始前的数据转换 hook（默认实现：恒等变换）

        可在子类中重写以实现测试时增强（TTA）

        Parameters
        ----------
        - `batch` : `TensorOrTensors`
            - 原始批次数据
        - `batch_idx` : `int`
            - 批次索引
        - `epoch_idx` : `Optional[int]`, optional
            - 当前 epoch 索引

        Returns
        -------
        - `TensorOrTensors`
            - 转换后的批次数据
        """
        return batch



"""
一个不依赖第三方 SDK 的简易 OpenAI LLM 接口实现
"""

import typing

import json
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, TypedDict, Union

from typing_extensions import Literal, NotRequired

# import warnings
from zyplib.import_tools import import_lazy
from zyplib.utils.print import print_debug, print_error

if TYPE_CHECKING:
    import requests
    import requests as RQ
else:
    requests = import_lazy('requests')

# ========================= OpenAI LLM Completion ========================= #


class ImageContent(TypedDict, total=False):
    """Represents image content in a message."""

    type: Literal['image_url']
    image_url: Dict[str, str]  # {"url": "..."}


class TextContent(TypedDict, total=False):
    """Represents text content in a message."""

    type: Literal['text']
    text: str


MessageContent = Union[str, List[Union[TextContent, ImageContent]]]
MessageRole = Literal['user', 'assistant', 'system']


class IChatOption(TypedDict, total=False):
    """Options for the LLM completion request."""

    temperature: NotRequired[float]
    top_p: NotRequired[float]
    max_tokens: NotRequired[int]
    frequency_penalty: NotRequired[float]
    presence_penalty: NotRequired[float]
    stop: NotRequired[Union[str, List[str]]]
    stream: NotRequired[bool]


class ILLMModel(TypedDict, total=False):
    """Represents configuration for a specific model API endpoint."""

    modelId: NotRequired[str]  # Model name sent to the API
    modelName: str  # User-facing model identifier; 通常情况下, 可以和 modelId 相同, 不过特殊情况下, 可能需要配置; 比如 Ark 平台使用的是 ID 化的模型，不方便查看
    url: str
    apiKey: str


class MessageItem(TypedDict, total=False):
    role: MessageRole
    content: MessageContent
    reasoning_content: NotRequired[str]  # 推理内容 (用于 o1 等推理模型)
    tool_call_id: NotRequired[str]  # 工具调用 ID
    tool_calls: NotRequired[List[Dict[str, Any]]]  # 工具调用


class CompletionResponse(MessageItem, total=False):
    """Represents a single message response from the LLM API."""

    usage: NotRequired[Dict[str, Any]]  # 使用统计信息
    finished: NotRequired[bool]  # 是否为最后一个 chunk


class StreamChunk(TypedDict, total=False):
    """Represents a single chunk from streaming response."""

    content: NotRequired[str]
    reasoning_content: NotRequired[str]  # 推理内容 (用于 o1 等推理模型)
    cumulative_content: NotRequired[str]  # 累积的完整内容
    cumulative_reasoning: NotRequired[str]  # 累积的推理内容
    usage: NotRequired[Dict[str, Any]]
    finished: NotRequired[bool]  # 是否为最后一个 chunk


def _copy_message_field(result: MessageItem, key: str, value: object) -> None:
    if key == 'role' and value in ('user', 'assistant', 'system'):
        result['role'] = typing.cast(MessageRole, value)
    elif key == 'content' and (isinstance(value, str) or isinstance(value, list)):
        result['content'] = typing.cast(MessageContent, value)
    elif key == 'tool_call_id' and isinstance(value, str):
        result['tool_call_id'] = value
    elif key == 'tool_calls' and isinstance(value, list):
        result['tool_calls'] = typing.cast(List[Dict[str, Any]], value)


def _adapt_input_message(
    input_data: Union[str, typing.Sequence[MessageItem]],
) -> List[MessageItem]:
    """
    Adapt input message to standard format and filter unsupported content based on model capabilities.

    Parameters
    ----------
    - `input_data` : `Union[str, List[MessageItem]]`
        - Input data, either a string or list of message items

    Returns
    ----------
    - `List[MessageItem]`
        - Standardized list of message items
    """
    messages: List[MessageItem] = []

    # Convert string input to message format
    if isinstance(input_data, str):
        messages = [{'role': 'user', 'content': input_data}]
    else:
        # Filter allowed fields and create clean messages
        allowed_fields = [
            'role',
            'content',
            'tool_call_id',
            'tool_calls',
            # 'reasoning_content',
        ]
        messages = []

        for item in input_data:
            result: MessageItem = {}
            for key, value in item.items():
                if key in allowed_fields:
                    _copy_message_field(result, key, value)
            messages.append(result)

    return messages


def _print_stream(chunk: 'StreamChunk'):
    print(chunk.get('content', ''), end='')


def openai_complete(
    prompt: Union[str, MessageItem, typing.Sequence[MessageItem]],
    model: Optional[ILLMModel] = None,
    options: Optional[IChatOption] = None,
    systemPrompt: Optional[str] = None,
    history: typing.Sequence[MessageItem] | None = None,
    proxy: Literal['system', 'none'] | str | None = None,
    return_type: Literal[
        'text', 'message_item', 'message_history', 'completion_response'
    ] = 'text',
    stream: bool = False,
    on_chunk_cb: Optional[Callable[[StreamChunk], None]] = _print_stream,
    chunk_cb_interval: int = 1,
) -> Union[str, MessageItem, List[MessageItem], CompletionResponse]:
    """
    Complete text using OpenAI-compatible API with optional streaming support.

    Supports reasoning models (e.g., o1-series) by automatically handling `reasoning_content`
    fields in the response.

    Parameters
    ----------
    - `prompt` : `Union[str, List[MessageItem]]`
        - The input prompt text or list of messages
    - `model` : `ILLMModel`, optional
        - Model configuration including URL and API key
    - `options` : `IChatOption`, optional
        - Chat completion options like temperature, max_tokens, etc.
    - `systemPrompt` : `str`, optional
        - System prompt to prepend to the conversation
    - `history` : `List[MessageItem]`, optional
        - Previous conversation history
    - `proxy` : `Literal['system', 'none'] | str`, optional
        - Proxy configuration, by default None
    - `return_type` : `Literal['text', 'message_item', 'message_history', 'completion_response']`, optional
        - Format of the return value, by default 'text'
    - `stream` : `bool`, optional
        - Enable streaming response, by default False
    - `on_chunk_cb` : `Callable[[StreamChunk], None]`, optional
        - Callback function for streaming chunks. Receives StreamChunk with content,
          reasoning_content, cumulative_content, and cumulative_reasoning fields, by default None

    Returns
    ----------
    - `Union[str, MessageItem, List[MessageItem], CompletionResponse]`
        - Response in the specified format. For reasoning models, includes formatted reasoning content.
    """
    options = options or {}

    # 1. Resolve the model identifier to get API details
    if not model:
        from zyplib._config import config

        api_url = config.OPENAI_BASE_URL
        api_key = config.OPENAI_KEY
        model_name_to_use = config.OPENAI_MODEL
    else:
        api_url = model.get('url')
        api_key = model.get('apiKey')
        model_name_to_use = model.get('modelId', model.get('modelName'))

    if not api_url or not model_name_to_use or not api_key:
        raise ValueError(
            'Model configuration is incomplete. Please provide a valid model with URL and API key.'
        )

    if not api_url or not model_name_to_use:
        raise RuntimeError(
            f"Resolved model configuration is incomplete for '{model}'. Missing URL or model name."
        )

    # api url trim last `/` if endsd with v1 / v2 v3 ... add /chat/completions
    if api_url.endswith('/'):
        api_url = api_url[:-1]  # Remove trailing slash
    if api_url.endswith('/chat/completions'):
        ...
    else:
        api_url = api_url + '/chat/completions'

    messages: List[MessageItem] = [] if history is None else list(history)
    if isinstance(prompt, str):
        messages.append({'role': 'user', 'content': prompt})
    elif isinstance(prompt, dict):
        # If prompt is a single MessageItem, add it directly
        messages.append(prompt)
    else:
        # If prompt is a list/sequence of MessageItem
        messages.extend(list(prompt))

    full_messages = _adapt_input_message(messages)

    if systemPrompt:
        # 独立处理系统提示; 防止 complete 返回List[MessageItem]中包含系统提示导致重复插入
        full_messages.insert(0, {'role': 'system', 'content': systemPrompt})

    # 2. Construct the request payload (OpenAI chat completions format)
    payload: Dict[str, Any] = {
        'model': model_name_to_use,
        'messages': full_messages,
        **options,
    }

    # 设置 stream 模式
    if stream:
        payload['stream'] = True
        # 启用 usage 数据返回
        payload['stream_options'] = {'include_usage': True}
    else:
        payload.pop('stream', None)

    # 3. Construct headers
    headers = {'Content-Type': 'application/json'}
    if api_key:  # Only add Authorization if an API key is provided
        headers['Authorization'] = f'Bearer {api_key}'

    print_debug(
        f'[LLM Request] POST {api_url}; {len(payload["messages"])} -> {model_name_to_use} (stream={stream})'
    )

    # 4. Make the API call using requests
    try:
        # Prepare session with proxy configuration
        session = requests.Session()

        if proxy is None or proxy == 'none':
            session.trust_env = False  # Disable system proxy
        elif proxy == 'system':
            session.trust_env = True  # Use system proxy
        elif isinstance(proxy, str):
            session.proxies = {'http': proxy, 'https': proxy}

        response = session.post(api_url, json=payload, headers=headers, stream=stream)

        # 5. Handle the response
        response_status = response.status_code

        if 200 <= response_status < 300:
            print_debug(f'[LLM Response] Status: {response_status}')

            if stream:
                return _handle_stream_response(
                    response, messages, return_type, on_chunk_cb, chunk_cb_interval
                )
            else:
                return _handle_normal_response(response, messages, return_type)
        else:
            response_text = response.text
            print_error(
                f'LLM API request failed with status {response_status}. Response: {response_text}'
            )
            raise RuntimeError(
                f'API request failed. Status: {response_status}, Response: {response_text[:500]}'
            )

    except requests.RequestException as e:
        print_error(f'Network error during API call: {e}')
        raise RuntimeError(f'Network error during API call: {e}') from e
    except Exception as e:
        print_error(f'An unexpected error occurred in openai_complete: {e}')
        raise RuntimeError(f'An unexpected error occurred: {e}') from e


def _handle_normal_response(
    response: 'RQ.Response', messages: typing.Sequence[MessageItem], return_type: str
) -> Union[str, MessageItem, List[MessageItem], CompletionResponse]:
    """Handle non-streaming response."""
    try:
        response_data = response.json()

        # Extract content and reasoning - adjust path based on actual API response structure
        message = response_data.get('choices', [{}])[0].get('message', {})
        content = message.get('content')
        reasoning = message.get('reasoning_content', '') or message.get('reasoning', '')

        if content is None:
            print_error(
                f'Failed to extract content from LLM response. Structure: {response_data}'
            )
            raise RuntimeError('LLM response format unexpected: Could not find content.')

        # Format content with reasoning
        # final_content = _format_content_with_reasoning(str(content), reasoning)

        print_debug(
            f'[LLM Response] Success, Content: {len(content)}, Reasoning: {len(reasoning)}'
        )

        # Create response message
        response_message: CompletionResponse = {
            'role': 'assistant',
            'content': content,
            'reasoning_content': reasoning,
            'finished': True,
        }
        usage = response_data.get('usage')
        if isinstance(usage, dict):
            response_message['usage'] = usage

        if return_type == 'text':
            return content
        elif return_type == 'message_item':
            return {
                'role': 'assistant',
                'content': content,
                'reasoning_content': reasoning,
            }
        elif return_type == 'message_history':
            return list(messages) + [
                {'role': 'assistant', 'content': content, 'reasoning_content': reasoning}
            ]
        elif return_type == 'completion_response':
            return response_message

        raise ValueError(f'Unsupported return_type: {return_type}')

    except json.JSONDecodeError as e:
        print_error(f'Failed to parse JSON response: {e}')
        raise RuntimeError(f'Failed to parse JSON response: {e}') from e


def _handle_stream_response(
    response: 'RQ.Response',
    messages: typing.Sequence[MessageItem],
    return_type: str,
    on_chunk_cb: Optional[Callable[[StreamChunk], None]],
    chunk_cb_interval: int = 1,
) -> Union[str, MessageItem, List[MessageItem], CompletionResponse]:
    """Handle streaming response."""
    cumulative_content = ''
    cumulative_reasoning = ''
    usage_info = None

    chunk_idx = 0  # 用于跟踪 chunk 调用次数

    try:
        for raw_chunk in response.iter_lines(decode_unicode=False):
            line = raw_chunk.decode('utf-8')  # 手动处理 UTF-8 解码
            if not line or not line.strip():
                continue

            if line.strip() == 'data: [DONE]':
                # 发送最终 chunk
                if on_chunk_cb:
                    final_chunk: StreamChunk = {
                        'content': '',
                        'cumulative_content': cumulative_content,
                        'cumulative_reasoning': cumulative_reasoning,
                        'finished': True,
                    }
                    if isinstance(usage_info, dict):
                        final_chunk['usage'] = usage_info
                    on_chunk_cb(final_chunk)
                break

            if line.startswith('data: '):
                try:
                    chunk_data = json.loads(line[6:])  # Remove 'data: ' prefix

                    # Handle usage information
                    if 'usage' in chunk_data:
                        usage_info = chunk_data['usage']

                    # Extract content from delta
                    choices = chunk_data.get('choices', [])
                    if choices:
                        delta = choices[0].get('delta', {})
                        chunk_content = delta.get('content', '')
                        chunk_reasoning = delta.get('reasoning_content', '') or delta.get(
                            'reasoning', ''
                        )

                        # 更新累积内容
                        if chunk_content:
                            cumulative_content += chunk_content
                        if chunk_reasoning:
                            cumulative_reasoning += chunk_reasoning

                        # 调用回调函数 (只有在有新内容时才调用)
                        if (
                            (chunk_content or chunk_reasoning)
                            and on_chunk_cb
                            and (chunk_idx % chunk_cb_interval == 0)
                        ):
                            stream_chunk: StreamChunk = {
                                'content': chunk_content,
                                'reasoning_content': chunk_reasoning,
                                'cumulative_content': cumulative_content,
                                'cumulative_reasoning': cumulative_reasoning,
                                'finished': False,
                            }
                            on_chunk_cb(stream_chunk)

                        chunk_idx += 1  # 增加 chunk 调用次数

                except json.JSONDecodeError as e:
                    print_error(f'Failed to parse stream chunk: {line} - {e}')
                    continue
                except Exception as e:
                    print_error(f'Error processing stream chunk: {e}')
                    continue

        print_debug(
            f'[LLM Stream] Complete, Content: {len(cumulative_content)}, Reasoning: {len(cumulative_reasoning)}'
        )

        # Create response message
        response_message: CompletionResponse = {
            'role': 'assistant',
            'content': cumulative_content,
            'reasoning_content': cumulative_reasoning,
            'finished': True,
        }
        if isinstance(usage_info, dict):
            response_message['usage'] = usage_info

        if return_type == 'text':
            return cumulative_content
        elif return_type == 'message_item':
            return {
                'role': 'assistant',
                'content': cumulative_content,
                'reasoning_content': cumulative_reasoning,
            }
        elif return_type == 'message_history':
            return list(messages) + [
                {
                    'role': 'assistant',
                    'content': cumulative_content,
                    'reasoning_content': cumulative_reasoning,
                }
            ]
        elif return_type == 'completion_response':
            return response_message

        raise ValueError(f'Unsupported return_type: {return_type}')

    except Exception as e:
        print_error(f'Error in stream response handling: {e}')
        raise RuntimeError(f'Error in stream response handling: {e}') from e
